import datetime
import logging
import random
import uuid
import warnings
from collections.abc import Generator

import kubernetes
import pytest

from geneva.config import override_config_kv
from geneva.runners.ray.raycluster import RayCluster, _HeadGroupSpec, _WorkerGroupSpec

kubernetes.config.load_kube_config()

logging.basicConfig(level=logging.INFO)
# integ test specific config overrides
override_config_kv(
    {
        "uploader.upload_dir": "gs://geneva-integ-test/zips",
        "job.checkpoint.mode": "object_store",
        "job.checkpoint.object_store.path": "gs://geneva-integ-test/checkpoints",
    }
)
# it's okay, we are in a test
warnings.filterwarnings(
    "ignore", "Using port forwarding for Ray cluster is not recommended for production"
)


@pytest.fixture(autouse=True, scope="session")
def k8s_core_api() -> kubernetes.client.CoreV1Api:
    """
    This fixture is used to create a CoreV1Api object for the test session.
    """
    return kubernetes.client.CoreV1Api()


@pytest.fixture(autouse=True)
def k8s_temp_service_account(
    k8s_core_api: kubernetes.client.CoreV1Api,
) -> Generator[str, None, None]:
    name = f"geneva-test-{uuid.uuid4().hex}"
    k8s_core_api.create_namespaced_service_account(
        namespace="geneva",
        body={
            "apiVersion": "v1",
            "kind": "ServiceAccount",
            "metadata": {
                "name": name,
                "namespace": "geneva",
            },
        },
    )
    yield name
    k8s_core_api.delete_namespaced_service_account(
        name=name,
        namespace="geneva",
        body=kubernetes.client.V1DeleteOptions(),
    )


@pytest.fixture(autouse=True, scope="session")
def geneva_k8s_service_account() -> str:
    """
    A preconfigured service account for the test session.
    This service account should have all the permissions needed to run the tests.
    """
    return "geneva-integ-test"


@pytest.fixture(autouse=True, scope="session")
def geneva_test_bucket() -> str:
    """
    A preconfigured service account for the test session.
    This service account should have all the permissions needed to run the tests.
    """
    return "gs://geneva-integ-test/data"


@pytest.fixture(autouse=True)
def standard_cluster(
    geneva_k8s_service_account: str,
) -> RayCluster:
    cluster_name = "geneva-integ-test"
    cluster_name += f"-{datetime.datetime.now().strftime('%Y-%m-%d-%H-%M')}"
    cluster_name += f"-{random.randint(0, 10000)}"

    return RayCluster(
        name=cluster_name,
        namespace="geneva",
        use_port_forward=True,
        head_group=_HeadGroupSpec(
            service_account=geneva_k8s_service_account,
            num_cpus=4,
            memory=4 * 4 * 1024**3,
        ),
        # allocate at least a single worker so the test runs faster
        # that we save time on waiting for the actor to start
        worker_groups=[
            _WorkerGroupSpec(
                name="worker",
                min_replicas=1,
                service_account=geneva_k8s_service_account,
                num_cpus=16,
                memory=4 * 16 * 1024**3,
            )
        ],
    )



# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: Copyright The Geneva Authors

from pyarrow import DataType


def datafusion_type_name(data_type: DataType) -> str:
    arrow_type_name = str(data_type)
    # see https://datafusion.apache.org/user-guide/sql/data_types.html
    # TODO: add more types. Note that we only support certain types in lance
    # https://github.com/lancedb/lance/blob/644213b9a63e2b143d62cda79e108df831bc5054/rust/lance-datafusion/src/planner.rs#L426-L441
    df_type_name = {
        "int8": "TINYINT",
        "uint8": "TINYINT UNSIGNED",
        "int16": "SMALLINT",
        "uint16": "SMALLINT UNSIGNED",
        "int32": "INT",
        "uint32": "INT UNSIGNED",
        "int64": "BIGINT",
        "uint64": "BIGINT UNSIGNED",
        "float32": "FLOAT",
        "float64": "DOUBLE",
        "string": "STRING",
        "binary": "BINARY",
        "boolean": "BOOLEAN",
    }.get(arrow_type_name)

    if df_type_name is None:
        raise ValueError(f"unsupported arrow type {arrow_type_name}")

    return df_type_name



# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: Copyright The Geneva Authors

"""Checkpoint Store for Geneva Pipeline"""

import abc
import logging
import os
import tempfile
from collections.abc import Iterator
from enum import Enum

import attrs
import pyarrow as pa
from lance.file import LanceFileReader, LanceFileWriter
from pyarrow import fs

from geneva.config import ConfigBase

_LOG = logging.getLogger(__name__)


class CheckpointStore(abc.ABC):
    """Abstract class for checkpoint store, which is used to store intermediate results
      of Geneva pipelines.

    It is implemented as a key-value store of :class:`pyarrow.RecordBatch` objects.

    This is a lighter weight version of collections.abc.MutableMapping
    where we don't expose length or deletion operations
    """

    @abc.abstractmethod
    def __contains__(self, item: str) -> bool:
        pass

    @abc.abstractmethod
    def __getitem__(self, item: str) -> pa.RecordBatch:
        pass

    @abc.abstractmethod
    def __setitem__(self, key: str, value: pa.RecordBatch) -> None:
        pass

    @abc.abstractmethod
    def list_keys(self, prefix: str = "") -> Iterator[str]:
        """List all the available keys for check point."""

    @classmethod
    def from_uri(cls, uri: str) -> "CheckpointStore":
        """Construct a CheckpointStore from a URI."""
        if uri == "memory":
            return InMemoryCheckpointStore()
        try:
            return LanceCheckpointStore(uri)
        except Exception as e:
            raise ValueError(f"Invalid checkpoint store uri: {uri}") from e


class InMemoryCheckpointStore(CheckpointStore):
    """In memory checkpoint store for testing purposes."""

    def __init__(self) -> None:
        self._store = {}

    def __contains__(self, item: str) -> bool:
        return item in self._store

    def __getitem__(self, item: str) -> pa.RecordBatch:
        return self._store[item]

    def __setitem__(self, key: str, value: pa.RecordBatch) -> None:
        self._store[key] = value

    def list_keys(self, prefix: str = "") -> Iterator[str]:
        for key in self._store:
            if key.startswith(prefix):
                yield key


class LanceCheckpointStore(CheckpointStore):
    def __init__(self, root: str) -> None:
        self.fs, self.path = fs.FileSystem.from_uri(root)  # type: ignore
        self.root = root

    def __contains__(self, key: str) -> bool:
        _LOG.debug("contains: %s", key)
        path = os.path.join(self.path, f"{key}.lance")
        info = self.fs.get_file_info(path)
        return info.type != fs.FileType.NotFound

    def __getitem__(self, key: str) -> pa.RecordBatch:
        _LOG.debug("get: %s", key)
        path = os.path.join(self.root, f"{key}.lance")
        reader = LanceFileReader(path)
        return reader.read_all().to_table().combine_chunks().to_batches()[0]

    def __setitem__(self, key: str, value: pa.RecordBatch) -> None:
        _LOG.debug("set: %s", key)
        path = os.path.join(self.root, f"{key}.lance")
        with LanceFileWriter(path, value.schema) as writer:
            writer.write_batch(value)

    def list_keys(self, prefix: str = "") -> Iterator[str]:
        _LOG.debug("list_keys: %s", prefix)
        selector = fs.FileSelector(os.path.join(self.path, prefix))
        for file in self.fs.get_file_info(selector):
            yield file.path.removeprefix(self.path).lstrip("/").rstrip(".lance")


class CheckpointMode(Enum):
    OBJECT_STORE = "object_store"

    # Store checkpoints in temporary files, for local development
    # and testing.
    # It can be shared between process, i.e., local ray actors.
    TEMPFILE = "tempfile"

    # for testing only
    IN_MEMORY = "in_memory"

    @staticmethod
    def from_str(s: str) -> "CheckpointMode":
        return CheckpointMode(s)


@attrs.define
class ObjectStoreCheckpointConfig(ConfigBase):
    path: str

    @classmethod
    def name(cls) -> str:
        return "object_store"

    def make(self) -> CheckpointStore:
        return LanceCheckpointStore(self.path)


@attrs.define
class CheckpointConfig(ConfigBase):
    mode: CheckpointMode = attrs.field(
        default=CheckpointMode.OBJECT_STORE, converter=CheckpointMode.from_str
    )

    object_store: ObjectStoreCheckpointConfig | None = attrs.field(default=None)

    @classmethod
    def name(cls) -> str:
        return "checkpoint"

    def make(self) -> CheckpointStore:
        match self.mode:
            case CheckpointMode.TEMPFILE:
                temp_dir = tempfile.mkdtemp()
                _LOG.info("Create checkpoint store on %s", temp_dir)
                return LanceCheckpointStore(temp_dir)
            case CheckpointMode.OBJECT_STORE:
                if self.object_store is None:
                    raise ValueError("CheckpointConfig::object_store is required")
                return LanceCheckpointStore(self.object_store.path)
            case CheckpointMode.IN_MEMORY:
                return InMemoryCheckpointStore()
            case _:
                raise ValueError(f"Unknown checkpoint mode {self.mode}")



# ruff: noqa: PERF203
# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: Copyright The Geneva Authors

# multi-process applier

import io
from collections.abc import Callable
from typing import Literal

import attrs
import multiprocess
import pyarrow as pa

from geneva.apply.applier import BatchApplier
from geneva.apply.task import MapTask, ReadTask
from geneva.debug.logger import ErrorLogger


def _buf_to_batch(
    buf: io.BytesIO,
    *,
    coalesce: bool = False,
) -> list[pa.RecordBatch] | pa.RecordBatch:
    """
    Convert a buffer to a record batch.
    """
    buf.seek(0)
    with pa.ipc.open_stream(buf) as f:
        t = f.read_all()
    if not coalesce:
        return t.to_batches()
    else:
        return t.combine_chunks().to_batches()[0]


def _batch_to_buf(
    batch: pa.RecordBatch,
) -> io.BytesIO:
    """
    Convert a record batch to a buffer.
    """
    buf = io.BytesIO()
    with pa.ipc.new_stream(buf, schema=batch.schema) as f:
        f.write_batch(batch)
    buf.seek(0)
    return buf


def _apply_with_stream_buf(
    apply: Callable[[pa.RecordBatch], pa.RecordBatch],
    buf: io.BytesIO,
) -> io.BytesIO:
    """
    Apply a function to a record batch using a stream buffer.
    """
    batch = _buf_to_batch(buf)

    out_buf = io.BytesIO()
    out_batches = [apply(batch) for batch in _buf_to_batch(buf)]
    with pa.ipc.new_stream(out_buf, schema=out_batches[0].schema) as f:
        for batch in out_batches:
            f.write_batch(batch)

    return out_buf


@attrs.define
class MultiProcessBatchApplier(BatchApplier):
    """
    A multi-process applier that applies a function to each element in the batch.
    """

    num_processes: int = attrs.field(validator=attrs.validators.ge(1))

    method: Literal["fork", "spawn"] = attrs.field(default="fork")

    def run(
        self,
        read_task: ReadTask,
        map_task: MapTask,
        error_logger: ErrorLogger,
    ) -> pa.RecordBatch:
        ctx = (
            multiprocess.context.ForkContext()
            if self.method == "fork"
            else multiprocess.context.SpawnContext()
        )

        with ctx.Pool(self.num_processes) as pool:
            # don't pull new batches until the previous ones are done
            # this way we reduce the number of batches in memory
            def _run_with_backpressure():  # noqa: ANN202
                futs = []
                seqs = []

                for seq, batch in enumerate(
                    read_task.to_batches(batch_size=map_task.batch_size())
                ):
                    # TODO: allow configuring the global batch size via config
                    seqs.append(seq)
                    data = _batch_to_buf(batch)

                    futs.append(
                        pool.apply_async(
                            _apply_with_stream_buf, args=(map_task.apply, data)
                        )
                    )
                    # don't start waiting till we have primed the queue
                    if len(futs) >= self.num_processes + 1:
                        seq = seqs.pop(0)
                        fut = futs.pop(0)
                        try:
                            yield _buf_to_batch(fut.get(), coalesce=True)
                        except Exception as e:
                            error_logger.log_error(e, read_task, seq)
                            raise e

                while futs:
                    seq = seqs.pop(0)
                    fut = futs.pop(0)
                    try:
                        yield _buf_to_batch(fut.get(), coalesce=True)
                    except Exception as e:
                        error_logger.log_error(e, read_task, seq)
                        raise e

            batches = list(_run_with_backpressure())

            if not batches:
                return pa.RecordBatch.from_arrays([])

            schema = batches[0].schema
            combined_table = pa.Table.from_batches(batches, schema)
            combined_table = combined_table.combine_chunks()
            combined_batch = combined_table.to_batches()[0]

            return combined_batch



# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: Copyright The Geneva Authors

import itertools
import logging
import random
from collections.abc import Callable, Iterator
from typing import TypeVar

import attrs
import lance
import more_itertools
import pyarrow as pa

from geneva.apply.applier import BatchApplier
from geneva.apply.simple import SimpleApplier
from geneva.apply.task import CopyTask, MapTask, ReadTask, ScanTask
from geneva.checkpoint import CheckpointStore
from geneva.debug.logger import ErrorLogger, NoOpErrorLogger
from geneva.table import TableReference

_LOG = logging.getLogger(__name__)


@attrs.define
class CheckpointingApplier:
    """
    Reads a read task and applies a map task to the data
    using a batch applier.

    The applier will checkpoint the output of the map task so that it can be
    resumed from the same point if the job is interrupted.
    """

    checkpoint_store: CheckpointStore = attrs.field()
    map_task: MapTask = attrs.field()

    error_logger: ErrorLogger = attrs.field(default=NoOpErrorLogger())
    batch_applier: BatchApplier = attrs.field(
        factory=SimpleApplier,
        converter=attrs.converters.default_if_none(factory=SimpleApplier),
    )

    @property
    def output_schema(self) -> pa.Schema:
        return self.map_task.output_schema()

    def _run(self, task: ReadTask) -> str:
        data_key = task.checkpoint_key()
        _LOG.debug("Running task %s", task)
        # track the batch sequence number so we can checkpoint any errors
        # when reproducing locally we can seek to the erroring batch quickly

        checkpoint_key = f"{data_key}:{self.map_task.checkpoint_key()}"
        if checkpoint_key in self.checkpoint_store:
            _LOG.info("Using cached result for %s", checkpoint_key)
            return checkpoint_key

        batch = self.batch_applier.run(
            task,
            self.map_task,
            error_logger=self.error_logger,
        )

        self.checkpoint_store[checkpoint_key] = batch

        return checkpoint_key

    def run(self, task: ReadTask) -> str:
        try:
            return self._run(task)
        except Exception as e:
            logging.exception("Error running task %s: %s", task, e)
            raise RuntimeError(f"Error running task {task}") from e

    def status(self, task: ReadTask) -> bool:
        data_key = task.checkpoint_key()
        return f"{data_key}:{self.map_task.checkpoint_key()}" in self.checkpoint_store


def _plan_read(
    uri: str,
    columns: list[str],
    *,
    read_version: int | None = None,
    batch_size: int = 512,
    filter: str | None = None,  # noqa: A002
    num_frags: int | None = None,
) -> Iterator[ReadTask]:
    """Make Plan for Reading Data from a Dataset"""
    dataset = lance.dataset(uri)
    if read_version is not None:
        dataset = dataset.checkout_version(read_version)

    for idx, frag in enumerate(dataset.get_fragments()):
        if num_frags is not None and idx >= num_frags:
            break
        frag_rows = frag.count_rows(filter=filter)
        for offset in range(0, frag_rows, batch_size):
            limit = min(batch_size, frag_rows - offset)
            yield ScanTask(
                uri=uri,
                version=read_version,
                columns=columns,
                frag_id=frag.fragment_id,
                offset=offset,
                limit=limit,
                filter=filter,
                with_row_address=True,
            )


T = TypeVar("T")  # Define type variable "T"


@attrs.define
class _LanceReadPlanIterator(Iterator[T]):
    it: Iterator[T]
    total: int

    def __iter__(self) -> Iterator[T]:
        return self

    def __next__(self) -> T:
        return next(self.it)

    def __len__(self) -> int:
        return self.total


def _num_tasks(
    *,
    uri: str,
    read_version: int | None = None,
    batch_size: int = 512,
) -> int:
    return sum(
        -(-frag.count_rows() // batch_size)
        for frag in (lance.dataset(uri, version=read_version)).get_fragments()
    )


T = TypeVar("T")


def _buffered_shuffle(it: Iterator[T], buffer_size: int) -> Iterator[T]:
    """Shuffle an iterator using a buffer of size buffer_size
    not perfectly random, but good enough for spreading out IO
    """
    # Initialize the buffer with the first buffer_size items from the iterator
    buffer = []
    # Fill the buffer with up to buffer_size items initially
    try:
        for _ in range(buffer_size):
            item = next(it)
            buffer.append(item)
    except StopIteration:
        pass

    while True:
        # Select a random item from the buffer
        index = random.randint(0, len(buffer) - 1)
        item = buffer[index]

        # Try to replace the selected item with a new one from the iterator
        try:
            next_item = next(it)
            buffer[index] = next_item
            # Yield the item AFTER replacing it in the buffer
            # this way the buffer is always contiguous so we can
            # simply yield the buffer at the end
            yield item
        except StopIteration:
            yield from buffer
            break


R = TypeVar("R")


def diversity_aware_shuffle(
    it: Iterator[T],
    key: Callable[[T], R],
    *,
    diversity_goal: int = 4,
    buffer_size: int = 1024,
) -> Iterator[T]:
    """A shuffle iterator that is aware of the diversity of the data
    being shuffled. The key function should return a value that is
    is used to determine the diversity of the data. The diversity_goal
    is the number of unique values that should be in the buffer at any
    given time. if the buffer is full, the items is yielded in a round-robin
    fashion. This is useful for shuffling tasks that are diverse, but

    This algorithm is bounded in memory by the buffer_size, so it is reasonably
    efficient for large datasets.
    """

    # NOTE: this is similar to itertools.groupby, but with a buffering limit

    buffer: dict[R, list[T]] = {}
    buffer_total_size = 0

    peekable_it = more_itertools.peekable(it)

    def _maybe_consume_from_iter() -> bool:
        nonlocal buffer_total_size
        item = peekable_it.peek(default=None)
        if item is None:
            return False
        key_val = key(item)
        if key_val not in buffer and len(buffer) < diversity_goal:
            buffer[key_val] = []
        else:
            return False

        # if the buffer still has room, add the item
        if buffer_total_size < buffer_size:
            buffer[key_val].append(item)
            buffer_total_size += 1
        else:
            return False

        next(peekable_it)
        return True

    while _maybe_consume_from_iter():
        ...

    production_counter = 0

    def _next_key() -> T | None:
        nonlocal buffer_total_size, production_counter
        if not buffer_total_size:
            return None

        # TODO: add warning about buffer size not big enough for diversity_goal
        buffer_slot = production_counter % len(buffer)
        key = next(itertools.islice(buffer.keys(), buffer_slot, buffer_slot + 1))
        assert key in buffer
        key_buffer = buffer[key]

        buffer_total_size -= 1
        item = key_buffer.pop(0)
        if not key_buffer:
            del buffer[key]

        # try to fill the removed buffer slot
        _maybe_consume_from_iter()
        production_counter += 1
        return item

    while (item := _next_key()) is not None:
        yield item


def plan_read(
    uri: str,
    columns: list[str],
    *,
    read_version: int | None = None,
    batch_size: int = 512,
    filter: str | None = None,  # noqa: A002
    shuffle_buffer_size: int = 0,
    task_shuffle_diversity: int | None = None,
    num_frags: int | None = None,
    **kwargs,
) -> Iterator[ReadTask]:
    """Make Plan for Reading Data from a Dataset"""
    it = _plan_read(
        uri,
        columns=columns,
        read_version=read_version,
        batch_size=batch_size,
        filter=filter,
        num_frags=num_frags,
    )
    # same as no shuffle
    if shuffle_buffer_size > 1 and task_shuffle_diversity is None:
        it = _buffered_shuffle(it, buffer_size=shuffle_buffer_size)
    elif task_shuffle_diversity is not None:
        buffer_size = max(4 * task_shuffle_diversity, shuffle_buffer_size)
        it = diversity_aware_shuffle(
            it,
            key=lambda task: task.checkpoint_key(),
            diversity_goal=task_shuffle_diversity,
            buffer_size=buffer_size,
        )

    return _LanceReadPlanIterator(
        it, _num_tasks(uri=uri, read_version=read_version, batch_size=batch_size)
    )


def _plan_copy(
    src: TableReference,
    dst: TableReference,
    columns: list[str],
    *,
    batch_size: int = 512,
) -> tuple[Iterator[CopyTask], int]:
    """Make Plan for Reading Data from a Dataset"""
    dst_dataset = dst.open().to_lance()
    num_tasks = 0
    for frag in dst_dataset.get_fragments():
        frag_rows = frag.count_rows()
        # ceil_div
        num_tasks += -(frag_rows // -batch_size)

    def task_gen() -> Iterator[CopyTask]:
        for frag in dst_dataset.get_fragments():
            frag_rows = frag.count_rows()
            for offset in range(0, frag_rows, batch_size):
                limit = min(batch_size, frag_rows - offset)
                yield CopyTask(
                    src=src,
                    dst=dst,
                    columns=columns,
                    frag_id=frag.fragment_id,
                    offset=offset,
                    limit=limit,
                )

    return (task_gen(), num_tasks)


def plan_copy(
    src: TableReference,
    dst: TableReference,
    columns: list[str],
    *,
    batch_size: int = 512,
    shuffle_buffer_size: int = 0,
    task_shuffle_diversity: int | None = None,
) -> Iterator[CopyTask]:
    (it, num_tasks) = _plan_copy(
        src,
        dst,
        columns,
        batch_size=batch_size,
    )
    # same as no shuffle
    if shuffle_buffer_size > 1 and task_shuffle_diversity is None:
        it = _buffered_shuffle(it, buffer_size=shuffle_buffer_size)
    elif task_shuffle_diversity is not None:
        buffer_size = max(4 * task_shuffle_diversity, shuffle_buffer_size)
        it = diversity_aware_shuffle(
            it,
            key=lambda task: task.checkpoint_key(),
            diversity_goal=task_shuffle_diversity,
            buffer_size=buffer_size,
        )

    return _LanceReadPlanIterator(it, num_tasks)



# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: Copyright The Geneva Authors

# tests for the zip packager

import re
import site
import subprocess
import uuid
from pathlib import Path

import pytest

from geneva.packager.autodetect import package_local_env_into_zips
from geneva.packager.zip import WorkspaceUnzipper, WorkspaceZipper

_MODULE_TEMPLATE = "print({})"


def _write_and_register_module(path: Path, content: str) -> None:
    with (path / "_geneva_zip_test.py").open("w") as f:
        f.write(content)
    site.addsitedir(path.as_posix())


def test_subprocess_can_not_import_without_zip_packaging(tmp_path: Path) -> None:
    uid = uuid.uuid4().hex
    _write_and_register_module(tmp_path, _MODULE_TEMPLATE.format(f'"{uid}"'))
    p = subprocess.Popen(
        "python -c 'import _geneva_zip_test'",
        shell=True,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )
    p.wait()
    stderr = p.stderr.read().decode()
    stdout = p.stdout.read().decode()
    assert p.returncode != 0

    assert stdout == ""
    assert any(
        re.compile(r"ModuleNotFoundError").match(line) for line in stderr.splitlines()
    )


def test_subprocess_can_import_with_zip_packaging(tmp_path: Path) -> None:
    uid = uuid.uuid4().hex
    _write_and_register_module(tmp_path, _MODULE_TEMPLATE.format(f'"{uid}"'))
    zipper = WorkspaceZipper(tmp_path, tmp_path)
    zip_path, _ = zipper.zip()

    content = f"""
from geneva.packager.zip import WorkspaceUnzipper
from pathlib import Path

unzipper = WorkspaceUnzipper()

unzipper.unzip(Path("{zip_path[0]}"))

import _geneva_zip_test
    """.strip()

    p = subprocess.Popen(
        f"python -c '{content}'",
        shell=True,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )
    p.wait()
    stderr = p.stderr.read().decode()
    stdout = p.stdout.read().decode()
    assert p.returncode == 0, f"stdout: {stdout}, stderr: {stderr}"

    assert stderr == ""
    assert stdout == f"{uid}\n"


def test_can_import_and_create_unzipper() -> None:
    WorkspaceUnzipper()


# SKIP because takes too much disk space
# move to integration tests https://github.com/lancedb/geneva/issues/122
@pytest.mark.skip
def test_autodetect(tmp_path: Path) -> None:
    package_local_env_into_zips(tmp_path)
    # TODO: check that the zips are created correctly



# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: Copyright The Geneva Authors

import itertools
from collections.abc import Callable
from pathlib import Path
from typing import Any, NamedTuple, cast

import lance
import pandas as pd
import pyarrow as pa
import pyarrow.compute as pc
import pytest

import geneva
from geneva import LanceCheckpointStore, connect, udf
from geneva.apply.multiprocess import MultiProcessBatchApplier
from geneva.apply.simple import SimpleApplier
from geneva.packager import DockerUDFPackager
from geneva.table import TableReference
from geneva.transformer import UDF

try:
    import ray

    from geneva.runners.ray.pipeline import (
        _simulate_write_failure,
        run_ray_add_column,
        run_ray_copy_table,
    )
except ImportError:
    import pytest

    pytest.skip("failed to import geneva.runners.ray", allow_module_level=True)


def make_new_ds(tbl_path: Path, fn) -> None:
    # create initial dataset with only column 'a'
    data = {"a": pa.array(range(256))}
    tbl = pa.Table.from_pydict(data)
    ds = lance.write_dataset(tbl, tbl_path, max_rows_per_file=16)

    # then add column 'b' using merge.  This is a separate commit from data
    # commits to keep column 'a' as a separate set of physical files from 'b'
    # which enables a separate commit from distributed execution to only
    # update 'b' with an efficient file replacement operation.
    new_frags = []
    new_schema = None
    for frag in ds.get_fragments():
        new_fragment, new_schema = frag.merge_columns(fn, columns=["a"])
        new_frags.append(new_fragment)

    assert new_schema is not None
    merge = lance.LanceOperation.Merge(new_frags, new_schema)
    lance.LanceDataset.commit(tbl_path, merge, read_version=ds.version)


class UDFTestConfig(NamedTuple):
    udf_fn: UDF
    recordbatch_schema_fn: Callable[[pa.RecordBatch], pa.RecordBatch]
    expected_recordbatch: dict[Any, Any]


def add_one_return_none(batch: pa.RecordBatch) -> pa.RecordBatch:
    return pa.RecordBatch.from_pydict(
        {"b": pa.array([None] * batch.num_rows, pa.int32())}
    )


# 0.1 cpu so we don't wait for provisioning in the tests
@udf(data_type=pa.int32(), batch_size=8, num_cpus=0.1)
def add_one(a) -> int:
    return a + 1


add_one_udftest = UDFTestConfig(
    add_one,
    add_one_return_none,
    {
        "a": list(range(256)),
        "b": [x + 1 for x in range(256)],
    },
)


def verify_run_ray_add_udf_column(
    tmp_path: Path, shuffle_config, udf: UDFTestConfig, batch_applier=None
) -> None:
    tbl_path = tmp_path / "foo.lance"
    make_new_ds(tbl_path, udf.recordbatch_schema_fn)
    tbl_ref = TableReference(db_uri=str(tmp_path), table_name="foo", version=None)

    ckp_store = LanceCheckpointStore(str(tmp_path / "ckp"))
    run_ray_add_column(
        tbl_ref,
        ["a"],
        {"b": udf.udf_fn},
        checkpoint_store=ckp_store,
        batch_applier=batch_applier,
        **shuffle_config,
    )

    ds = lance.dataset(tbl_path)
    assert ds.to_table().to_pydict() == udf.expected_recordbatch


@pytest.mark.parametrize(
    ("shuffle_config", "batch_applier"),
    [
        (
            {
                "batch_size": batch_size,
                "shuffle_buffer_size": shuffle_buffer_size,
                "task_shuffle_diversity": task_shuffle_diversity,
            },
            batch_applier,
        )
        for (
            batch_size,
            shuffle_buffer_size,
            task_shuffle_diversity,
            batch_applier,
        ) in itertools.product(
            [4, 16],
            [7],
            [3],
            [MultiProcessBatchApplier(num_processes=4), SimpleApplier(), None],
        )
    ],
)
def test_run_ray_add_column(tmp_path: Path, shuffle_config, batch_applier) -> None:
    verify_run_ray_add_udf_column(
        tmp_path, shuffle_config, add_one_udftest, batch_applier
    )


@pytest.mark.parametrize(
    "batch_applier", [None, SimpleApplier(), MultiProcessBatchApplier(num_processes=4)]
)
def test_run_ray_add_column_write_fault(tmp_path: Path, batch_applier) -> None:  # noqa: PT019
    tbl_path = tmp_path / "foo.lance"
    make_new_ds(tbl_path, add_one_return_none)
    tbl_ref = TableReference(db_uri=str(tmp_path), table_name="foo", version=None)

    ckp_store = LanceCheckpointStore(str(tmp_path / "ckp"))

    with _simulate_write_failure(True):
        run_ray_add_column(
            tbl_ref,
            ["a"],
            {"b": add_one},
            checkpoint_store=ckp_store,
            batch_applier=batch_applier,
        )

    ds = lance.dataset(tbl_path)
    assert ds.to_table().to_pydict() == {
        "a": list(range(256)),
        "b": [x + 1 for x in range(256)],
    }


def test_run_ray_add_column_with_deletes(tmp_path: Path) -> None:  # noqa: PT019
    tbl_path = tmp_path / "foo.lance"
    make_new_ds(tbl_path, add_one_return_none)
    tbl_ref = TableReference(db_uri=str(tmp_path), table_name="foo", version=None)

    ds = lance.dataset(tbl_path)
    ds.delete("a % 2 == 1")

    ckp_store = LanceCheckpointStore(str(tmp_path / "ckp"))

    run_ray_add_column(tbl_ref, ["a"], {"b": add_one}, checkpoint_store=ckp_store)

    ds = lance.dataset(tbl_path)
    assert ds.to_table().to_pydict() == {
        "a": list(range(0, 256, 2)),
        "b": [x + 1 for x in range(0, 256, 2)],
    }


struct_type = pa.struct([("rpad", pa.string()), ("lpad", pa.string())])


def struct_udf_return_none(batch: pa.RecordBatch) -> pa.RecordBatch:
    return pa.RecordBatch.from_pydict(
        {"b": pa.array([None] * batch.num_rows, struct_type)}
    )


@udf(data_type=struct_type, batch_size=8, num_cpus=0.1)
def struct_udf(a: pa.Array) -> pa.Array:  # is the output type correct?
    rpad = pc.ascii_rpad(pc.cast(a, target_type="string"), 4, padding="0")
    lpad = pc.ascii_lpad(pc.cast(a, target_type="string"), 4, padding="0")
    return pc.make_struct(rpad, lpad, field_names=["rpad", "lpad"])


ret_struct_udftest = UDFTestConfig(
    struct_udf,
    struct_udf_return_none,
    {
        "a": list(range(256)),
        "b": [{"lpad": f"{x:04d}", "rpad": f"{x}0000"[:4]} for x in range(256)],
    },
)

default_shuffle_config = {
    "batch_size": 1,
    "shuffle_buffer_size": 3,
    "task_shuffle_diversity": None,
}


def test_run_ray_add_column_ret_struct(tmp_path: Path) -> None:
    verify_run_ray_add_udf_column(tmp_path, default_shuffle_config, ret_struct_udftest)


vararray_type = pa.list_(pa.int64())


def vararray_udf_return_none(batch: pa.RecordBatch) -> pa.RecordBatch:
    return pa.RecordBatch.from_pydict(
        {"b": pa.array([[]] * batch.num_rows, vararray_type)}
    )


@udf(data_type=vararray_type, batch_size=8, num_cpus=0.1)
def vararray_udf(a: pa.Array) -> pa.Array:  # is the output type correct?
    # [ [], [1], [2,2], [3,3,3] ... ]
    arr = [[val] * val for val in a.to_pylist()]
    b = pa.array(arr, type=pa.list_(pa.int64()))
    return b


ret_vararray_udftest = UDFTestConfig(
    vararray_udf,
    vararray_udf_return_none,
    {
        "a": list(range(256)),
        "b": [[x] * x for x in range(256)],
    },
)


def test_run_ray_add_column_ret_vararray(tmp_path: Path) -> None:
    verify_run_ray_add_udf_column(
        tmp_path, default_shuffle_config, ret_vararray_udftest
    )


@udf(data_type=vararray_type, batch_size=8, num_cpus=0.1)
class StatefulVararrayUDF(Callable):
    def __init__(self) -> None:
        self.state = 0

    def __call__(self, a: pa.Array) -> pa.Array:  # is the output type correct?
        # [ [], [1], [2,2], [3,3,3] ... ]
        arr = [[val] * val for val in a.to_pylist()]
        b = pa.array(arr, type=pa.list_(pa.int64()))
        return b


ret_vararray_stateful_udftest = UDFTestConfig(
    StatefulVararrayUDF,
    vararray_udf_return_none,
    {
        "a": list(range(256)),
        "b": [[x] * x for x in range(256)],
    },
)


@pytest.mark.skip(
    reason="this test requires extra variables but the ray init fixture seems"
    " to deadlock"
)
def test_run_ray_add_column_ret_vararray_stateful(tmp_path: Path) -> None:
    verify_run_ray_add_udf_column(
        tmp_path, default_shuffle_config, ret_vararray_stateful_udftest
    )


nested_type = pa.struct([("lpad", pa.string()), ("array", pa.list_(pa.int64()))])


def nested_udf_return_none(batch: pa.RecordBatch) -> pa.RecordBatch:
    return pa.RecordBatch.from_pydict(
        {"b": pa.array([{"lpad": "", "array": []}] * batch.num_rows, nested_type)}
    )


@udf(data_type=nested_type, batch_size=8, num_cpus=0.1)
def nested_udf(a: pa.Array) -> pa.Array:
    # [ { lpad:"0000", array:[] } , {lpad:"0001", array:[1]},
    #   { lpad:"0002", array:[2,2]}, ... ]

    lpad = pc.ascii_lpad(pc.cast(a, target_type="string"), 4, padding="0")
    arr = [[val] * val for val in a.to_pylist()]
    array = pa.array(arr, type=pa.list_(pa.int64()))

    return pc.make_struct(lpad, array, field_names=["lpad", "array"])


ret_nested_udftest = UDFTestConfig(
    nested_udf,
    nested_udf_return_none,
    {
        "a": list(range(256)),
        "b": [{"lpad": f"{val:04d}", "array": [val] * val} for val in range(256)],
    },
)


def test_run_ray_add_column_ret_nested(tmp_path: Path) -> None:
    verify_run_ray_add_udf_column(tmp_path, default_shuffle_config, ret_nested_udftest)


@pytest.mark.skip(reason="do we need to support relative paths?")
def test_relative_path(tmp_path: Path, monkeypatch) -> None:
    # Make sure this ray instance uses the tmp_path as CURDIR
    ray.shutdown()
    monkeypatch.chdir(tmp_path)

    packager = DockerUDFPackager(
        # use prebuilt image tag so we don't have to build the image
        prebuilt_docker_img="test-image:latest"
    )
    db = geneva.connect("./db", packager=packager)

    # create a basic table
    tbl = pa.Table.from_pydict({"id": [1, 2, 3, 4, 5, 6]})
    table = db.create_table("table1", tbl)

    @udf(data_type=pa.int64())
    def double_id(id: int):  # noqa A002
        return id * 2

    table.add_columns(
        {"id2": double_id},
    )

    schema = table.schema
    field = schema.field("id2")
    assert field.metadata[b"virtual_column.udf_name"] == b"double_id"

    # At this time, "id2" is a null column
    assert table.to_arrow().combine_chunks() == pa.Table.from_pydict(
        {"id": [1, 2, 3, 4, 5, 6], "id2": [None] * 6},
        schema=pa.schema(
            [
                pa.field("id", pa.int64()),
                pa.field("id2", pa.int64(), True),
            ]
        ),
    )

    # uses local ray to execute UDF and populate "id2"
    table.materialize("id2")

    df = table.to_arrow().to_pandas()
    assert df.equals(
        pd.DataFrame({"id": [1, 2, 3, 4, 5, 6], "id2": [2, 4, 6, 8, 10, 12]})
    )


def test_ray_materialized_view(tmp_path: Path) -> None:
    packager = DockerUDFPackager(prebuilt_docker_img="test-image:latest")
    db = connect(tmp_path, packager=packager)
    tbl = pa.Table.from_pydict({"video_uri": ["a", "b", "c", "d", "e", "f"]})
    table = db.create_table("table", tbl)

    @udf(data_type=pa.binary())
    def load_video(video_uri: pa.Array) -> pa.Array:
        videos = [str(i).encode("utf-8") for i in video_uri]
        return cast("pa.Array", pa.array(videos))

    load_video = cast("UDF", load_video)

    view_table = (
        table.search(None)
        .shuffle(seed=42)
        .select(
            {
                "video_uri": "video_uri",
                "video": load_video,
            }
        )
        .create_materialized_view(db, "table_view")
    )

    ckp_store = LanceCheckpointStore(str(tmp_path / "ckp"))

    run_ray_copy_table(view_table.get_reference(), packager, ckp_store)

    view_table.checkout_latest()
    assert view_table.to_arrow() == pa.Table.from_pydict(
        {
            "__source_row_id": [3, 2, 5, 4, 1, 0],
            "__is_set": [False] * 6,
            "video_uri": ["d", "c", "f", "e", "b", "a"],
            "video": [b"d", b"c", b"f", b"e", b"b", b"a"],
        }
    )



# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: Copyright The Geneva Authors

# a module for automatically inspecting path and capturing the local env

import base64
import logging
import site
import tempfile
from concurrent.futures import ThreadPoolExecutor, as_completed
from pathlib import Path

import emoji

from geneva.packager.uploader import Uploader
from geneva.packager.zip import WorkspaceZipper
from geneva.tqdm import tqdm

_LOG = logging.getLogger(__name__)


def _maybe_pyproject() -> Path | None:
    """
    Inspect the current working directory and return the path to the
    source root if it's in a python project.
    """

    # don't resolve because we could be inside a link
    cwd = Path.cwd().absolute()
    assert cwd.root == "/", "cwd is not absolute"

    root = Path("/")
    cur = cwd

    while not list(cur.glob("pyproject.toml")) and (cur := cur.parent) != root:
        ...

    # TODO: use the packaging tool configured in pyproject.toml
    # to determine the source root
    if list(cur.glob("pyproject.toml")):
        return cur / "src"

    return None


def _maybe_src() -> Path | None:
    """
    Inspect the current working directory and return the path to the
    source root if it's in a python project.
    """

    # don't resolve because we could be inside a link
    cwd = Path.cwd().absolute()
    assert cwd.root == "/", "cwd is not absolute"

    root = Path("/")
    cur = cwd

    while not list(cur.glob("src")) and (cur := cur.parent) != root:
        ...

    if list(cur.glob("src")):
        return cur / "src"

    return None


def _maybe_python_repo() -> Path | None:
    """
    Inspect the current working directory and return the path to the
    source root if it's in a python project.
    """

    return _maybe_pyproject() or _maybe_src()


def _find_workspace() -> Path:
    """
    Inspect the current working directory and return the path to the
    source root if it's in a python project. If we can not find a
    python source root, return the current working directory.
    """

    return _maybe_python_repo() or Path.cwd()


def get_paths_to_package() -> list[Path]:
    """
    Inspect the current working directory and return the path to the
    source root if it's in a python project. If we can not find a
    python source root, return the current working directory.
    """

    paths = sorted(
        {
            Path(p).resolve().absolute()
            for p in [
                _find_workspace(),
                Path.cwd(),
                *site.getsitepackages(),
            ]
        }
    )

    _LOG.info("found paths to package: %s", paths)

    return paths


def package_local_env_into_zips(output_dir: Path | str) -> list[list[Path]]:
    """
    Package the local environment into zip files.

    return a list of the paths to the zip files.
    """

    paths = get_paths_to_package()

    zips = []

    for p in paths:
        if not p.exists():
            _LOG.warning("path %s does not exist", p)
            continue

        if not p.is_dir():
            _LOG.warning("path %s is not a directory", p)
            continue

        _LOG.info("packaging %s", p)

        path_b32 = base64.b32encode(p.as_posix().encode()).decode()

        zip_path, _ = WorkspaceZipper(
            p,
            output_dir,
            file_name=f"{path_b32}.zip",
        ).zip()

        zips.append(zip_path)

    return zips


def upload_local_env(
    *,
    zip_output_dir: Path | str | None = None,
    uploader: Uploader | None = None,
) -> list[list[str]]:
    zip_output_dir = zip_output_dir or tempfile.mkdtemp()

    zips = package_local_env_into_zips(zip_output_dir)
    uploader = uploader or Uploader.get()

    sizes = []
    with ThreadPoolExecutor(max_workers=8) as executor:
        res = []
        for zip_paths in zips:
            res.append([])
            for zip_path in zip_paths:
                fut = executor.submit(
                    uploader.upload,
                    zip_path,
                )
                sizes.append(zip_path.stat().st_size)
                res[-1].append(fut)

        with tqdm(
            total=sum(sizes), unit="B", unit_scale=True, unit_divisor=1024, smoothing=0
        ) as pbar:
            pbar.set_description(emoji.emojize(":rocket: uploading zips"))
            for i in range(len(res)):
                uploaded_uris = []
                for p in as_completed(res[i]):
                    pbar.update(sizes.pop(0))
                    uploaded_uris.append(p.result())
                res[i] = uploaded_uris

    return res



# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: Copyright The Geneva Authors

import hashlib
import json
import logging
from collections.abc import Iterable, Iterator
from datetime import timedelta
from functools import cached_property
from pathlib import Path
from typing import Any, Literal

import attrs
import lance
import lancedb
import numpy as np
import pyarrow as pa
from lancedb.common import VECTOR_COLUMN_NAME
from lancedb.index import IndexConfig
from lancedb.query import Query as LanceQuery
from lancedb.table import IndexStatistics
from lancedb.table import LanceTable as LanceLocalTable
from lancedb.table import Table as LanceTable
from pyarrow.fs import FileSystem, LocalFileSystem

# Python 3.10 compatibility
from typing_extensions import Never, override  # noqa: UP035

from geneva.checkpoint import InMemoryCheckpointStore, LanceCheckpointStore
from geneva.db import Connection, connect
from geneva.query import GenevaQueryBuilder
from geneva.transformer import UDF

_LOG = logging.getLogger(__name__)


@attrs.define(order=True)
class TableReference:
    db_uri: str
    table_name: str
    version: int | None

    def open(self) -> "Table":
        try:
            # Not clear about when I need a lance checkpoint store vs arrow
            # fs checkpoint store
            ckp_store = LanceCheckpointStore(str(Path(self.db_uri) / "ckp"))
        except Exception:
            # this is dirty
            ckp_store = InMemoryCheckpointStore()
        return connect(self.db_uri, checkpoint=ckp_store).open_table(
            self.table_name, version=self.version
        )


class Table(LanceTable):
    """Table in Geneva.

    A Table is a Lance dataset
    """

    def __init__(
        self,
        conn: Connection,
        name: str,
        *,
        version: int | None = None,
        storage_options: dict[str, str] | None = None,
        index_cache_size: int | None = None,
        **kwargs,
    ) -> None:
        self._conn_uri = conn.uri
        self._name = name

        self._conn = conn

        base_uri = conn._uri.removesuffix("/")
        self._uri = f"{base_uri}/{name}.lance"
        self._version: int | None = version
        self._index_cache_size = index_cache_size
        self._storage_options = storage_options

        # Load table
        self._ltbl  # noqa

    def __repr__(self) -> str:
        return f"<Table {self._name}>"

    # TODO: This annotation sucks
    def __reduce__(self):  # noqa: ANN204
        return (self.__class__, (self._conn, self._name))

    def get_reference(self) -> TableReference:
        return TableReference(
            db_uri=self._conn.uri, table_name=self._name, version=self._version
        )

    def get_fragments(self) -> list[lance.LanceFragment]:
        return self.to_lance().get_fragments()

    @cached_property
    def _ltbl(self) -> lancedb.table.Table:
        # remote db, open table directly
        if self._conn_uri.startswith("db://"):
            tbl = self._conn._connect.open_table(self._name)
            if self._version:
                tbl.checkout(self._version)
            return tbl

        return self._conn._connect.open_table(self.name)

    @property
    def name(self) -> str:
        """Get the name of the table."""
        return self._name

    @property
    def version(self) -> int:
        """Get the current version of the table"""
        return self._ltbl.version

    @property
    def schema(self) -> pa.Schema:
        """The Arrow Schema of the Table."""
        return self._ltbl.schema

    @property
    def uri(self) -> str:
        return self._uri

    @property
    def embedding_functions(self) -> Never:
        raise NotImplementedError("Embedding functions are not supported.")

    def add(
        self,
        data,
        mode: str = "append",
        on_bad_vectors: str = "error",
        fill_value: float = 0.0,
    ) -> None:
        self._ltbl.add(
            data,
            mode=mode,
            on_bad_vectors=on_bad_vectors,
            fill_value=fill_value,
        )

    def checkout(self, version: int) -> None:
        self._version = version
        self._ltbl.checkout(version)

    def checkout_latest(self) -> None:
        self._ltbl.checkout_latest()

    def add_columns(
        self, transforms: dict[str, str | UDF | tuple[UDF, list[str]]], *args, **kwargs
    ) -> None:
        """
        Add columns or udf based columns to the Geneva table.

        Parameters
        ----------
        transforms : dict[str, str | UDF | tuple[UDF, list[str]]]
            The key is the column name to add and the value is a
            specification of the column type/value.
            * If the spec is a string, it is expected to be a datafusion
              sql expression. (e.g "cast(null as string)"))
            * If the spec is a UDF, a virtual column is added with input
              columns inferred from the UDF's argument names.
            * If the spec is a tuple, the first element is a UDF and the
              second element is a list of input column names.

        Raises
        ------
        ValueError

        """
        # handle basic columns
        basic_cols = {k: v for k, v in transforms.items() if isinstance(v, str)}
        if len(basic_cols) > 0:
            self._ltbl.add_columns(basic_cols, *args)

        # handle UDF virtual columns
        udf_cols = {k: v for k, v in transforms.items() if not isinstance(v, str)}
        for k, v in udf_cols.items():
            if isinstance(v, UDF):
                # infer column names from udf arguments
                udf = v
                self._add_virtual_columns(
                    {k: udf}, *args, input_columns=udf.input_columns, **kwargs
                )
            else:
                # explicitly specify input columns
                (udf, cols) = v
                self._add_virtual_columns({k: udf}, *args, input_columns=cols, **kwargs)

    def _add_virtual_columns(
        self,
        mapping: dict[str, UDF],  # this breaks the non udf mapping
        *args,
        input_columns: list[str] | None = None,
        **kwargs,
    ) -> None:
        """
        This is an internal method and not intended to be called directly.

        Add udf based virtual columns to the Geneva table.
        """

        if len(mapping) != 1:
            raise ValueError("Only one UDF is supported for now.")

        _LOG.info("Adding column: udf=%s", mapping)
        col_name = next(iter(mapping))
        udf = mapping[col_name]

        if not isinstance(udf, UDF):
            # Stateful udfs are implemenated as Callable classses, and look
            # like partial functions here.  Instantiate to get the return
            # data_type annotations.
            udf = udf()
        self._ltbl.add_columns(pa.field(col_name, udf.data_type))
        self._configure_virtual_column(col_name, udf, input_columns)

        if kwargs.get("materialize", False):
            kwargs.pop("materialize")
            self.materialize(col_name, **kwargs)

    def materialize(
        self,
        col_name,
        *,
        input_columns: list[str] | None = None,
        udf: UDF | None = None,
        **kwargs,
    ) -> None:
        if udf is None:
            from geneva.runners.ray.__main__ import fetch_udf

            field = self._ltbl.schema.field(col_name)
            input_columns = json.loads(
                field.metadata.get(b"virtual_column.udf_inputs", "null")
            )
            udf_spec = fetch_udf(self, col_name)
            udf = self._conn._packager.unmarshal(udf_spec)

        if udf is None or input_columns is None:
            raise ValueError(
                "udf and input_columns must be specified if the udf is not persisted"
            )

        from geneva.runners.ray.pipeline import run_ray_add_column

        checkpoint_store = self._conn._checkpoint_store
        run_ray_add_column(
            self.get_reference(),
            input_columns,
            {col_name: udf},
            checkpoint_store=checkpoint_store,
            **kwargs,
        )

        # updates came from an external writer, so get the latest version.
        self._ltbl.checkout_latest()

    def alter_columns(self, *alterations: Iterable[dict[str, Any]], **kwargs) -> None:
        basic_column_alterations = []
        for alter in alterations:
            if "virtual_column" in alter:
                if "path" not in alter:
                    raise ValueError("path is required to alter virtual virtual_column")
                if not isinstance(alter["virtual_column"], UDF):
                    raise ValueError("virtual_column must be a UDF")

                col_name = alter["path"]
                udf = alter["virtual_column"]

                input_cols = alter.get("input_columns", None)

                if input_cols is None:
                    input_cols = udf.input_columns

                self._configure_virtual_column(col_name, udf, input_cols)

                if kwargs.get("materialize", False):
                    self.materialize(col_name)

            else:
                basic_column_alterations.append(alter)

        if len(basic_column_alterations) > 0:
            self._ltbl.alter_columns(*basic_column_alterations)

    def _configure_virtual_column(
        self,
        col_name: str,
        udf: UDF,
        input_cols: list[str] | None,
    ) -> None:
        """
        Configure a column to be a virtual column for the given UDF.

        This procedure includes:
        - Packaging the UDF
        - Uploading the UDF to the dataset
        - Updating the field metadata to include the UDF information

        Note that the column should already exist on the table.
        """
        udf_spec = self._conn._packager.marshal(udf)

        # upload the UDF to the dataset URL
        if not isinstance(self._ltbl, LanceLocalTable):
            raise TypeError(
                "adding udf column is currently only supported for local tables"
            )

        # upload the packaged UDF to some location inside the dataset:
        ds = self.to_lance()
        fs, root_uri = FileSystem.from_uri(ds.uri)
        checksum = hashlib.sha256(udf_spec.udf_payload).hexdigest()
        udf_location = f"_udfs/{checksum}"

        # TODO -- only upload the UDF if it doesn't exist
        if isinstance(fs, LocalFileSystem):
            # Object storage filesystems like GCS and S3 will create the directory
            # automatically, but local filesystem will not, so we create explicitly
            fs.create_dir(f"{root_uri}/_udfs")

        with fs.open_output_stream(f"{root_uri}/{udf_location}") as f:
            f.write(udf_spec.udf_payload)

        field_metadata = {
            "virtual_column": "true",
            "virtual_column.udf_backend": udf_spec.backend,
            "virtual_column.udf_name": udf_spec.name,
            "virtual_column.udf": "_udfs/" + checksum,
            "virtual_column.udf_inputs": json.dumps(input_cols),
        }

        # Add the column metadata:
        self._ltbl.replace_field_metadata(col_name, field_metadata)

    def create_index(
        self,
        metric: str = "L2",
        num_partitions: int | None = None,
        num_sub_vectors: int | None = None,
        vector_column_name: str = VECTOR_COLUMN_NAME,
        replace: bool = True,
        accelerator=None,
        index_cache_size=None,
        *,
        index_type: Literal[
            "IVF_FLAT",
            "IVF_PQ",
            "IVF_HNSW_SQ",
            "IVF_HNSW_PQ",
        ] = "IVF_PQ",
        num_bits: int = 8,
        max_iterations: int = 50,
        sample_rate: int = 256,
        m: int = 20,
        ef_construction: int = 300,
    ) -> None:
        """Create Vector Index"""
        self._ltbl.create_index(
            metric,
            num_partitions or 256,
            num_sub_vectors or 96,
            vector_column_name,
            replace,
            accelerator,
            index_cache_size,
            index_type=index_type,
            num_bits=num_bits,
            max_iterations=max_iterations,
            sample_rate=sample_rate,
            m=m,
            ef_construction=ef_construction,
        )

    @override
    def create_fts_index(
        self,
        field_names: str | list[str],
        *,
        ordering_field_names: str | list[str] | None = None,
        replace: bool = False,
        writer_heap_size: int | None = None,
        tokenizer_name: str | None = None,
        with_position: bool = True,
        base_tokenizer: Literal["simple", "raw", "whitespace"] = "simple",
        language: str = "English",
        max_token_length: int | None = 40,
        lower_case: bool = True,
        stem: bool = False,
        remove_stop_words: bool = False,
        ascii_folding: bool = False,
        **_kwargs,
    ) -> None:
        self._ltbl.create_fts_index(
            field_names,
            ordering_field_names=ordering_field_names,
            replace=replace,
            writer_heap_size=writer_heap_size,
            tokenizer_name=tokenizer_name,
            with_position=with_position,
            base_tokenizer=base_tokenizer,
            language=language,
            max_token_length=max_token_length,
            lower_case=lower_case,
            stem=stem,
            remove_stop_words=remove_stop_words,
            ascii_folding=ascii_folding,
            use_tantivy=False,
        )

    @override
    def create_scalar_index(
        self,
        column: str,
        *,
        replace: bool = True,
        index_type: Literal["BTREE", "BITMAP", "LABEL_LIST"] = "BTREE",
    ) -> None:
        self._ltbl.create_scalar_index(
            column,
            replace=replace,
            index_type=index_type,
        )

    @override
    def _do_merge(self, other: "Table", on: list[str], how: str) -> Never:
        raise NotImplementedError("Merging tables is not supported.")

    @override
    def _execute_query(
        self,
        query: LanceQuery,
        batch_size: int | None = None,
    ) -> pa.RecordBatchReader:
        return self._ltbl._execute_query(query, batch_size=batch_size)

    def list_versions(self) -> list[dict[str, Any]]:
        return self._ltbl.list_versions()

    @override
    def cleanup_old_versions(
        self,
        older_than: timedelta | None = None,
        *,
        delete_unverified=False,
    ) -> "lance.CleanupStats":
        return self._ltbl.cleanup_old_versions(
            older_than,
            delete_unverified=delete_unverified,
        )

    def to_batches(self, batch_size: int | None = None) -> Iterator[pa.RecordBatch]:
        from .query import Query

        if isinstance(self._ltbl, Query):
            return self._ltbl.to_batches(batch_size)
        return self.to_lance().to_batches(batch_size)

    def search(
        self,
        query: list | pa.Array | pa.ChunkedArray | np.ndarray | None = None,
        vector_column_name: str | None = None,
        query_type: Literal["vector", "fts", "hybrid", "auto"] = "auto",
        ordering_field_name: str | None = None,
        fts_columns: str | list[str] | None = None,
    ) -> GenevaQueryBuilder:
        if (
            query is not None
            or vector_column_name is not None
            or fts_columns is not None
        ):
            raise ValueError("Geneva search only implemented for plain queries")
        return GenevaQueryBuilder(self)

    @override
    def drop_columns(self, columns: Iterable[str]) -> None:
        self._ltbl.drop_columns(columns)

    @override
    def to_arrow(self) -> pa.Table:
        return self._ltbl.to_arrow()

    @override
    def count_rows(self, filter: str | None = None) -> int:
        return self._ltbl.count_rows(filter)

    @override
    def update(
        self,
        where: str | None = None,
        values: dict | None = None,
        *,
        values_sql: dict[str, str] | None = None,
    ) -> None:
        self._ltbl.update(where, values, values_sql=values_sql)

    @override
    def delete(self, where: str) -> None:
        self._ltbl.delete(where)

    @override
    def list_indices(self) -> Iterable[IndexConfig]:
        return self._ltbl.list_indices()

    @override
    def index_stats(self, index_name: str) -> IndexStatistics | None:
        return self._ltbl.index_stats(index_name)

    @override
    def optimize(
        self,
        *,
        cleanup_older_than: timedelta | None = None,
        delete_unverified: bool = False,
    ) -> None:
        return self._ltbl.optimize(
            cleanup_older_than=cleanup_older_than,
            delete_unverified=delete_unverified,
        )

    @override
    def compact_files(self) -> None:
        self._ltbl.compact_files()

    @override
    def restore(self, *args, **kwargs) -> None:
        self._ltbl.restore(*args, **kwargs)

    # TODO: This annotation sucks
    def take_blobs(self, indices: list[int] | pa.Array, column: str):  # noqa: ANN201
        return self.to_lance().take_blobs(indices, column)

    def to_lance(self) -> lance.LanceDataset:
        return self._ltbl.to_lance()

    def uses_v2_manifest_paths(self) -> bool:
        return self._ltbl.uses_v2_manifest_paths()

    def migrate_v2_manifest_paths(self) -> None:
        return self._ltbl.migrate_v2_manifest_paths()

    def _analyze_plan(self, query: lancedb.query.Query) -> str:
        return self._ltbl._analyze_plan(query)

    def _explain_plan(
        self, query: lancedb.query.Query, verbose: bool | None = False
    ) -> str:
        return self._ltbl._explain_plan(query, verbose=verbose)



import argparse

import ray
from lancedb.table import LanceTable as LanceLocalTable
from pyarrow.fs import FileSystem

import geneva
from geneva import LanceCheckpointStore
from geneva.packager import DockerUDFPackager, UDFSpec
from geneva.runners.ray.pipeline import run_ray_add_column
from geneva.table import Table


def run_ray_job() -> None:
    parser = argparse.ArgumentParser()
    parser.add_argument("--db_uri", type=str, required=True)
    parser.add_argument("--table_name", type=str, required=True)
    parser.add_argument("--checkpoint_store", type=str, required=True)
    parser.add_argument("--column", type=str, required=True)
    args = parser.parse_args()

    conn = geneva.connect(args.db_uri)
    table = conn.open_table(args.table_name)

    udf_spec = fetch_udf(table, args.column)

    # TODO: maybe the class here should be infered from the backend arg
    packager = DockerUDFPackager(runtime_backend="ray")
    udf = packager.unmarshal(udf_spec)

    if udf.input_columns is None:
        raise ValueError("UDF must have input columns for add_column job")

    checkpoint_store = LanceCheckpointStore(args.checkpoint_store)
    run_ray_add_column(
        table.get_reference(),
        udf.input_columns,
        {
            args.column: udf,
        },
        checkpoint_store,
        batch_size=8,
    )


def ds_uri(table: Table) -> str:
    if not isinstance(table._ltbl, LanceLocalTable):
        raise ValueError(
            f"Table {table} is not a local table. Currently only local tables are supported."  # noqa E501
        )
    ds = table._ltbl.to_lance()
    return ds.uri


def fetch_udf(table: Table, column_name: str) -> UDFSpec:
    schema = table._ltbl.schema
    field = schema.field(column_name)
    if field is None:
        raise ValueError(f"Column {column_name} not found in table {table}")

    udf_path = metadata_value("virtual_column.udf", field.metadata)
    fs, root_uri = FileSystem.from_uri(ds_uri(table))
    udf_payload = fs.open_input_file(f"{root_uri}/{udf_path}").read()

    udf_name = metadata_value("virtual_column.udf_name", field.metadata)
    udf_backend = metadata_value("virtual_column.udf_backend", field.metadata)

    return UDFSpec(
        name=udf_name,
        backend=udf_backend,
        udf_payload=udf_payload,
    )


def metadata_value(key: str, metadata: dict[bytes, bytes]) -> str:
    value = metadata.get(key.encode("utf-8"))
    if value is None:
        raise ValueError(f"Metadata key {key} not found in metadata {metadata}")
    return value.decode("utf-8")


if __name__ == "__main__":
    ray.init()
    run_ray_job()



import pytest

from geneva.runners.ray.raycluster import RayCluster, _WorkerGroupSpec


def test_service_account_does_not_exist() -> None:
    cluster = RayCluster(
        name="not-used",
        namespace="geneva",
        use_port_forward=True,
        # allocate at least a single worker so the test runs faster
        # that we save time on waiting for the actor to start
        worker_groups=[
            _WorkerGroupSpec(
                name="worker", min_replicas=1, service_account="does-not-exist"
            )
        ],
    )
    with pytest.raises(
        ValueError, match="Service account does-not-exist does not exist"
    ):
        cluster._validate()


def test_service_account_not_enough_permission(
    k8s_temp_service_account: str,
) -> None:
    cluster = RayCluster(
        name="not-used",
        namespace="geneva",
        use_port_forward=True,
        # allocate at least a single worker so the test runs faster
        # that we save time on waiting for the actor to start
        worker_groups=[
            _WorkerGroupSpec(
                name="worker",
                min_replicas=1,
                service_account=k8s_temp_service_account,
            )
        ],
    )
    with pytest.raises(
        ValueError,
        match=f"Service account {k8s_temp_service_account} does not"
        " have the required permission:",
    ):
        cluster._validate()


def test_service_account_is_valid(
    geneva_k8s_service_account: str,
) -> None:
    cluster = RayCluster(
        name="not-used",
        namespace="geneva",
        use_port_forward=True,
        # allocate at least a single worker so the test runs faster
        # that we save time on waiting for the actor to start
        worker_groups=[
            _WorkerGroupSpec(
                name="worker",
                min_replicas=1,
                service_account=geneva_k8s_service_account,
            )
        ],
    )
    cluster._validate()



# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: Copyright The Geneva Authors

# A simple port forwarder for talking to ray head
# this is for testing only, and should NEVER be used in production

import logging
import select
import socket
import threading

import attrs
import kubernetes
import kubernetes.stream.ws_client
from typing_extensions import Self

from geneva.runners.ray.raycluster import RayCluster

_LOG = logging.getLogger(__name__)


@attrs.define
class _Flag:
    val: bool
    lock: threading.Lock = attrs.field(factory=threading.Lock, init=False)

    @classmethod
    def true(cls) -> Self:
        return cls(True)

    def __bool__(self) -> bool:
        with self.lock:
            return self.val

    def set(self, val: bool) -> None:
        with self.lock:
            self.val = val


@attrs.define
class PortForward:
    """
    A simple port forwarder for talking to a k8s pod

    This forwarder binds and listens on a local port. Any traffic
    sent to this port is forwarded to the specified pod and port.

    Note: when local_port is 0, a random port will be allocated by the OS
    and the local port will be set to that value. User is responsible
    for checking the local port value after the forwarder is started.
    """

    pod_name: str
    namespace: str
    port: int

    local_port: int = 0

    alive: _Flag = attrs.field(factory=_Flag.true, init=False)

    threads: list[threading.Thread] = attrs.field(factory=list, init=False)

    proxy_start_barrier: threading.Barrier = attrs.field(init=False)

    proxy_listener: socket.socket = attrs.field(init=False)

    @classmethod
    def to_head_node(
        cls,
        cluster: RayCluster,
    ) -> Self:
        pod = cluster.head_node_pod

        # TODO: allocate a random local port
        return cls(
            pod_name=pod.metadata.name,
            namespace=cluster.namespace,
            port=10001,
        )

    def _start_proxy(
        self,
        pod_socket: socket.socket,
        client_socket: socket.socket,
    ) -> None:
        _LOG.info("Starting proxy from pod to client")
        proxy_thread = threading.Thread(
            target=_proxy,
            args=(pod_socket, client_socket, self.alive),
        )
        self.threads.extend([proxy_thread])
        proxy_thread.start()

        _LOG.info("Proxy started")

    def _start_listener_loop(self) -> None:
        _LOG.info("Starting listener loop")
        core_api = kubernetes.client.CoreV1Api()

        proxy_listener = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        proxy_listener.bind(("localhost", self.local_port))
        if self.local_port == 0:
            self.local_port = proxy_listener.getsockname()[1]
        proxy_listener.listen()

        self.proxy_listener = proxy_listener

        self.proxy_start_barrier.wait()

        while self.alive:
            _LOG.info("Waiting for client connection")
            try:
                client_socket, _ = proxy_listener.accept()
            except OSError:
                # socket is closed, exit
                _LOG.info("Listener socket closed")
                break
            client_socket.setblocking(True)

            forward: kubernetes.stream.ws_client.PortForward = (
                kubernetes.stream.portforward(
                    core_api.connect_get_namespaced_pod_portforward,
                    self.pod_name,
                    self.namespace,
                    ports=f"{self.port}",
                )
            )

            pod_socket = forward.socket(self.port)
            pod_socket.setblocking(True)

            self._start_proxy(pod_socket, client_socket)

    def __enter__(self) -> Self:
        _LOG.info("Starting port forward")
        self.alive.set(True)
        self.proxy_start_barrier = threading.Barrier(2)
        listener_thread = threading.Thread(
            target=self._start_listener_loop,
        )
        self.threads.append(listener_thread)
        listener_thread.start()

        self.proxy_start_barrier.wait()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        _LOG.info("Stopping port forward")
        self.alive.set(False)

        # not a clean shutdown, but we don't care
        self.proxy_listener.shutdown(socket.SHUT_RDWR)

        for thread in self.threads:
            thread.join(timeout=1)

        self.threads.clear()
        _LOG.info("Port forward stopped")


def _proxy(
    s1: socket.socket,
    s2: socket.socket,
    alive=True,
    *,
    buffer_size: int = 4096,
) -> None:
    _LOG.info(f"Proxying between {s1} and {s2}")
    while alive:
        for s in select.select([s1, s2], [], [])[0]:
            data = s.recv(buffer_size)
            {s1: s2, s2: s1}[s].sendall(data)



import uuid

import pyarrow as pa

import geneva
from geneva.runners.ray.raycluster import RayCluster


# use a random version to force checkpoint to invalidate
@geneva.udf(num_cpus=1, version=uuid.uuid4().hex)
def plus_one(a: int) -> int:
    return a + 1


def test_ray_add_column_pipeline(
    geneva_test_bucket: str,
    standard_cluster: RayCluster,
) -> None:
    conn = geneva.connect(geneva_test_bucket)
    table_name = uuid.uuid4().hex
    table = conn.create_table(
        table_name,
        pa.Table.from_pydict({"a": pa.array(range(10240))}),
    )
    with standard_cluster:
        table.add_columns(
            {"b": plus_one},
            materialize=True,
            batch_size=32,
            concurrency=32,
        )

    assert table.to_arrow() == pa.Table.from_pydict(
        {"a": pa.array(range(10240)), "b": pa.array(range(1, 10241))}
    )
    conn.drop_table(table_name)



from dataclasses import dataclass, field
from heapq import heappop, heappush
from typing import Generic, TypeVar

T = TypeVar("T")


@dataclass(order=True)
class _QueueItem(Generic[T]):
    position: int
    size: int
    item: T = field(compare=False)


class SequenceQueue(Generic[T]):
    """A queue that maintains order based on sequence positions.

    Items can be inserted with their position and size, and the queue will maintain
    them in order. The pop() method will only return an item if the next
    expected position is available. When an item is popped, the next position
    advances by the item's size.

    Example:
        >>> q = SequenceQueue[int]()
        >>> q.put(2, 2, "second")  # position 2, size 2
        >>> q.put(0, 1, "first")   # position 0, size 1
        >>> q.put(4, 1, "third")   # position 4, size 1
        >>> q.pop()  # Returns "first" (advances position by 1)
        >>> q.pop()  # Returns "second" (advances position by 2)
        >>> q.pop()  # Returns "third" (advances position by 1)
    """

    def __init__(self) -> None:
        self._heap = []
        self._next_position = 0

    def put(self, position: int, size: int, item: T) -> None:
        """Insert an item with its position and size into the queue.

        Args:
            position: The sequence position of the item
            size: The size of the item (how much to advance next_position)
            item: The item to insert
        """
        heappush(self._heap, _QueueItem(position, size, item))

    def pop(self) -> T | None:
        """Pop the next item in sequence if available.

        Returns:
            The next item in sequence if available, None otherwise.
            An item is only returned if it matches the next expected position.
            When an item is popped, next_position advances by the item's size.
        """
        if not self._heap:
            return None

        next_item = self._heap[0]
        if next_item.position == self._next_position:
            heappop(self._heap)
            self._next_position += next_item.size
            return next_item.item
        return None

    def peek(self) -> T | None:
        """Peek at the next item in sequence without removing it.

        Returns:
            The next item in sequence if available and ready, None otherwise.
        """
        if not self._heap:
            return None

        next_item = self._heap[0]
        if next_item.position == self._next_position:
            return next_item.item
        return None

    def is_empty(self) -> bool:
        """Check if the queue is empty.

        Returns:
            True if the queue is empty, False otherwise.
        """
        return len(self._heap) == 0 or self._heap[0].position > self._next_position

    def next_position(self) -> int:
        """Get the next expected position.

        Returns:
            The next position that will be returned by pop()
        """
        return self._next_position



# ruff: noqa: F821

# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: Copyright The Geneva Authors

# For managing and launching Ray Cluster

import abc
import base64
import contextlib
import functools
import getpass
import json
import logging
import platform
import re
import sys
import tempfile
import time
import warnings
from collections import Counter
from collections.abc import Generator
from typing import Any

import attrs
import kubernetes
import kubernetes.client.exceptions
import pyarrow as pa
import ray

# do global config init
from kubernetes import config as kube_config
from kubernetes.client.api import core_v1_api, scheduling_v1_api

import geneva
from geneva.config import ConfigBase
from geneva.packager.autodetect import upload_local_env

kube_config.load_config()

_LOG = logging.getLogger(__name__)

GENEVA_RAY_HEAD_NODE = "geneva.lancedb.com/ray-head"
GENEVA_RAY_CPU_NODE = "geneva.lancedb.com/ray-worker-cpu"
GENEVA_RAY_GPU_NODE = "geneva.lancedb.com/ray-worker-gpu"


@attrs.define
class _RayClusterConfig(ConfigBase):
    user: str = attrs.field(
        converter=attrs.converters.default_if_none(default=getpass.getuser())
    )

    namespace: str = attrs.field(
        converter=attrs.converters.default_if_none(default="geneva")
    )

    @classmethod
    def name(cls) -> str:
        return "raycluster"

    def cluster_name(self) -> str:
        if self.user:
            return self.user
        _LOG.info("Using the current OS user name as the cluster name")
        return getpass.getuser()


def _size_converter(value: int | str) -> int:
    if isinstance(value, int):
        return value
    suffixes = {
        "K": 1000,
        "M": 1000**2,
        "G": 1000**3,
        "T": 1000**4,
        "Ki": 1024,
        "Mi": 1024**2,
        "Gi": 1024**3,
        "Ti": 1024**4,
    }
    match = re.match(r"(?P<value>\d+)(?P<unit>[KMGT]i?)", value)
    value = int(match.group("value"))
    unit = suffixes[match.group("unit")]

    return value * unit


class _ValidationVisitable(abc.ABC):
    @abc.abstractmethod
    def _validate(self, visitor: "_ValidationVisitor") -> None:
        """
        Validate at cluster construction time if the definition is valid
        """


@attrs.define(kw_only=True, slots=False)
class _ResourceMixin:
    num_cpus: int = attrs.field(default=1, validator=attrs.validators.gt(0))
    memory: int = attrs.field(
        converter=_size_converter,
        default=2 * (1024**3),
        validator=attrs.validators.gt(0),
    )
    num_gpus: int = attrs.field(default=0, validator=attrs.validators.ge(0))

    arm: bool = attrs.field(default=platform.processor() == "aarch64")

    @property
    def _resources(self) -> dict:
        resource = {
            "requests": {
                "cpu": self.num_cpus,
                "memory": self.memory,
            },
            "limits": {
                "cpu": self.num_cpus,
                "memory": self.memory,
            },
        }

        if self.num_gpus:
            resource["requests"]["nvidia.com/gpu"] = self.num_gpus
            resource["limits"]["nvidia.com/gpu"] = self.num_gpus

        return resource


@attrs.define(kw_only=True, slots=False)
class _ServiceAccountMixin(_ValidationVisitable):
    service_account: str | None = attrs.field(default=None)

    @property
    def _service_account(self) -> dict[str, str] | None:
        if self.service_account is None:
            return None
        return {
            "serviceAccountName": self.service_account,
        }

    def _validate(self, visitor: "_ValidationVisitor") -> None:
        visitor.visit_service_account(self)


@attrs.define(kw_only=True, slots=False)
class _PriorityClassMixin(_ValidationVisitable):
    priority_class: str | None = attrs.field(default=None)

    @property
    def _priority_class(self) -> dict[str, str] | None:
        if self.priority_class is None:
            return None
        return {
            "priorityClassName": self.priority_class,
        }

    def _validate(self, visitor: "_ValidationVisitor") -> None:
        visitor.visit_priority_class(self)


@attrs.define(kw_only=True, slots=False)
class _RayVersionMixin:
    ray_version: str = attrs.field(init=False)
    """
    The version of Ray to use for the cluster. Auto detected from the Ray
    package version in the current environment.
    """

    @ray_version.default
    def _default_ray_version(self) -> str:
        return ray.__version__


@attrs.define(kw_only=True, slots=False)
class _PythonVersionMixin:
    python_version: str = attrs.field(init=False)
    """
    The major.minor version of Python to use for the cluster.
    Auto detected from the python version in the current environment.
    """

    @python_version.default
    def _default_python_version(self) -> str:
        return f"{sys.version_info.major}.{sys.version_info.minor}"


def get_ray_image(
    version: str, python_version: str, *, gpu: bool = False, arm: bool = False
) -> str:
    py_version = python_version.replace(".", "")
    image = f"rayproject/ray:{version}-py{py_version}"
    if gpu:
        image += "-gpu"
    if arm:
        image += "-aarch64"
    return image


@attrs.define(kw_only=True, slots=False)
class _ImageMixin(_RayVersionMixin, _PythonVersionMixin, _ResourceMixin):
    # set a dummpy default so the generated __init__
    # gets the correct signature
    image: str = attrs.field(default="_DEFUALT_IMAGE")

    def __attrs_post_init__(self) -> None:
        if self.image == "_DEFUALT_IMAGE":
            self.image = get_ray_image(
                self.ray_version,
                self.python_version,
                gpu=self.num_gpus > 0,
                arm=self.arm,
            )


@attrs.define(kw_only=True, slots=False)
class _MountsMixin:
    volumes: dict[str, dict] = attrs.field(default={})
    """
    Volumes to attach to the worker Pod.

    The key is the name of the volume and the value is the volume specification.
    """

    mounts: list[tuple[str, str]] = attrs.field(default=[])
    """
    The list of mounts to attach to the worker containers.
    """

    @mounts.validator
    def _validate_mounts(self, attribute: str, value: list[tuple[str, str]]) -> None:
        paths = set()
        for name, path in value:
            if name not in self.volumes:
                raise ValueError(f"Volume {name} not found in volumes")
            paths.add(path)

        if len(paths) != len(value):
            dups = [
                item
                for item, count in Counter(path for _, path in value).items()
                if count > 1
            ]
            raise ValueError(f"Duplicate mount paths: {dups}")

    @property
    def mounts_definition(self) -> list[dict[str, str]]:
        return [{"name": volume, "mountPath": path} for volume, path in self.mounts]

    @property
    def volume_definition(self) -> list[dict[str, str]]:
        return [{"name": volume, **config} for volume, config in self.volumes.items()]


class _PodSpec(
    _ImageMixin,
    _ResourceMixin,
    _MountsMixin,
    _ServiceAccountMixin,
    _PriorityClassMixin,
    _ValidationVisitable,
):
    def _validate(self, visitor: "_ValidationVisitor") -> None:
        visitor.visit_service_account(self)
        visitor.visit_priority_class(self)


@attrs.define(kw_only=True)
class _HeadGroupSpec(_PodSpec):
    node_selector: dict[str, str] = attrs.field(default={"_PLACEHOLDER": "true"})

    def __attrs_post_init__(self) -> None:
        super().__attrs_post_init__()
        if self.node_selector == {"_PLACEHOLDER": "true"}:
            self.node_selector = {GENEVA_RAY_HEAD_NODE: ""}

    @property
    def _ports(self) -> list[dict]:
        return [
            {
                "containerPort": 10001,
                "name": "client",
                "protocol": "TCP",
            },
            {
                "containerPort": 8265,
                "name": "dashboard",
                "protocol": "TCP",
            },
            {
                "containerPort": 6379,
                "name": "gsc-server",
                "protocol": "TCP",
            },
        ]

    @property
    def definition(self) -> dict:
        definition = {
            "rayStartParams": {
                # do not schedule tasks onto the head node this prevents cluster
                # crashes due to other tasks killing the head node
                "num-cpus": "0"
            },
            "template": {
                "spec": {
                    **(self._priority_class or {}),
                    **(self._service_account or {}),
                    "containers": [
                        {
                            "name": "ray-head",
                            "image": self.image,
                            "imagePullPolicy": "IfNotPresent",
                            "resources": self._resources,
                            "ports": self._ports,
                            "volumeMounts": self.mounts_definition,
                        }
                    ],
                    "volumes": self.volume_definition,
                    "nodeSelector": self.node_selector,
                }
            },
        }

        return definition


@attrs.define(kw_only=True)
class _WorkerGroupSpec(_PodSpec):
    """
    A worker group specification for a Ray cluster.
    """

    name: str = "worker"

    node_selector: dict[str, str] = attrs.field(default={"_PLACEHOLDER": "true"})

    min_replicas: int = attrs.field(
        default=0,
        validator=attrs.validators.ge(0),
    )
    max_replicas: int = attrs.field(
        default=100,
    )

    @max_replicas.validator
    def _validate_max_replicas(self, attribute: str, value: int) -> None:
        if value == 0:
            raise ValueError("max_replicas must be greater than 0")

        if value < self.min_replicas:
            raise ValueError(
                f"max_replicas ({value}) must be greater than or",
                f" equal to min_replicas ({self.min_replicas})",
            )

    def __attrs_post_init__(self) -> None:
        super().__attrs_post_init__()
        if self.node_selector == {"_PLACEHOLDER": "true"}:
            if self.num_gpus > 0:
                self.node_selector = {GENEVA_RAY_GPU_NODE: ""}
            else:
                self.node_selector = {GENEVA_RAY_CPU_NODE: ""}

    @property
    def definition(self) -> dict:
        return {
            "groupName": self.name,
            "minReplicas": self.min_replicas,
            "maxReplicas": self.max_replicas,
            "rayStartParams": {
                "num-cpus": str(self.num_cpus),
            },
            "template": {
                "spec": {
                    **(self._priority_class or {}),
                    **(self._service_account or {}),
                    "containers": [
                        {
                            "name": "ray-worker",
                            "image": self.image,
                            "imagePullPolicy": "IfNotPresent",
                            "resources": self._resources,
                            "volumeMounts": self.mounts_definition,
                            "env": [
                                {
                                    "name": "RAY_memory_usage_threshold",
                                    "value": "0.9",  # Adjust the threshold as needed
                                },
                                {
                                    "name": "RAY_memory_monitor_refresh_ms",
                                    "value": "0",  # Set to 0 disables the auto-kill
                                },
                            ],
                        }
                    ],
                    "volumes": self.volume_definition,
                    "nodeSelector": self.node_selector,
                }
            },
        }


@attrs.define(kw_only=True, slots=False, frozen=False)
class RayCluster(_RayVersionMixin, _ValidationVisitable):
    """
    A Ray cluster specification.

    This is also a context manager for managing a Ray cluster.
    This context manager will apply the Ray cluster definition to the
    Kubernetes cluster when entering the context and delete the Ray
    cluster from the Kubernetes cluster when exiting the context.

    When entering the context, ray.init will be called with the address of the
    Ray cluster head node. When exiting the context, ray.shutdown will be
    called to shutdown the Ray cluster.

    Example:
        >>> from geneva.runners.ray.raycluster import RayCluster, cluster
        >>> head_group = _HeadGroupSpec(image="rayproject/ray:latest")
        >>> worker_group = _WorkerGroupSpec(name="worker", image="rayproject/ray")
        >>> with RayCluster(
        ...     name="test-cluster",
        ...     head_group=head_group,
        ...     worker_groups=[worker_group],
        ... ):
        ...     print("Ray cluster is running")
        Ray cluster is running
    """

    name: str = attrs.field()
    """
    The name of the Ray cluster. This name is used for deduplication of Ray
    clusters in the Kubernetes cluster. When a Ray cluster with the same name
    already exists, we will not create a new one.

    TODO: add a recreate=True option to force the recreation of the cluster.
    """

    @name.default
    def _default_name(self) -> str:
        config = _RayClusterConfig.get()
        return f"geneva-{config.user}"

    namespace: str = attrs.field()
    """
    The namespace of the Ray cluster. This is the namespace in which the Ray
    cluster will be created in the Kubernetes cluster.

    By default, we use `geneva` as the namespace.
    """

    @namespace.default
    def _default_namespace(self) -> str:
        config = _RayClusterConfig.get()
        return config.namespace

    head_group: _HeadGroupSpec = attrs.field(factory=_HeadGroupSpec)
    """
    The head group specification for the Ray cluster.
    """

    worker_groups: list[_WorkerGroupSpec] = attrs.field(
        factory=lambda: [_WorkerGroupSpec()]
    )
    """
    The worker group specifications for the Ray cluster.
    """

    zip_output_dir: str = attrs.field(default=tempfile.mkdtemp())
    """
    The directory to store the zipped environment files.
    """

    use_port_forward: bool = attrs.field(default=False)
    """
    Whether to use port forwarding to connect to the Ray cluster.

    This should only be used for testing as the port forwarding
    server is in-process and can not provide high performance.
    """

    @property
    def _autoscaler_options(self) -> dict:
        """
        The autoscaler options for the Ray cluster.
        """
        # TODO: allow customization of the autoscaler options
        return {
            "env": [],
            "envFrom": [],
            "idleTimeoutSeconds": 60,
            "imagePullPolicy": "IfNotPresent",
            "resources": {
                "requests": {
                    "cpu": "1",
                    "memory": "1Gi",
                },
                "limits": {
                    "cpu": "1",
                    "memory": "1Gi",
                },
            },
            "upscalingMode": "Default",
        }

    @property
    def spec(self) -> dict:
        """
        The Ray cluster specification.

        This can be used as part of RayJob for configuring the Ray cluster.
        """
        return {
            "enableInTreeAutoscaling": True,
            "autoscalerOptions": self._autoscaler_options,
            "rayVersion": self.ray_version,
            "headGroupSpec": self.head_group.definition,
            "workerGroupSpecs": [worker.definition for worker in self.worker_groups],
        }

    @property
    def definition(self) -> dict:
        """
        The Ray cluster definition.

        This is the full definition of the Ray cluster, including the name and
        autoscaler options. This can be used to create the Ray cluster in the
        Kubernetes cluster via a CRD.
        """
        return {
            "apiVersion": "ray.io/v1",
            "kind": "RayCluster",
            "metadata": {
                "name": self.name,
                "namespace": self.namespace,
            },
            "spec": self.spec,
        }

    def _has_existing_cluster(self, api: kubernetes.client.CustomObjectsApi) -> bool:
        try:
            api.get_namespaced_custom_object(
                group="ray.io",
                version="v1",
                namespace=self.namespace,
                plural="rayclusters",
                name=self.name,
            )
        except kubernetes.client.exceptions.ApiException as e:
            if e.status == 404:
                return False
            raise e
        return True

    def _wait_for_cluster(self, api: kubernetes.client.CustomObjectsApi) -> Any:
        while True:
            # TODO: add wait for the Ray cluster to be ready
            result = api.get_namespaced_custom_object(
                group="ray.io",
                version="v1",
                namespace=self.namespace,
                plural="rayclusters",
                name=self.name,
            )
            assert isinstance(result, dict)

            if "status" not in result:
                _LOG.info("Waiting for the Ray cluster to be ready")
                time.sleep(1)
                continue

            status = result["status"]
            assert isinstance(status, dict)

            if "head" not in status:
                _LOG.info("Waiting for the head node to be ready")
                time.sleep(1)
                continue

            head = status["head"]
            assert isinstance(head, dict)

            if "podIP" not in head:
                _LOG.info("Waiting for the head node IP address")
                time.sleep(1)
                continue

            _LOG.info("Ray cluster is ready")
            return result

    def _wait_for_head_node(self, api: core_v1_api.CoreV1Api, pod_name: str) -> Any:
        while True:
            pod = api.read_namespaced_pod(name=pod_name, namespace=self.namespace)
            if pod.status.phase != "Running":
                _LOG.info("Waiting for the head node to be running")
                time.sleep(1)
                continue

            _LOG.info("Head node is running")
            return pod

    @property
    def head_node_pod(self) -> Any:
        crd_api = kubernetes.client.CustomObjectsApi()
        cluster = self._wait_for_cluster(crd_api)
        pod_name = cluster["status"]["head"]["podName"]
        core_api = core_v1_api.CoreV1Api()
        pod = self._wait_for_head_node(core_api, pod_name)
        return pod

    def _expose_node_port(self, api: core_v1_api.CoreV1Api) -> None:
        _LOG.info("Exposing head node on a node port")
        # create a service to expose the head node
        api.create_namespaced_service(
            namespace=self.namespace,
            body={
                "apiVersion": "v1",
                "kind": "Service",
                "metadata": {
                    "name": f"{self.name}-head",
                    "namespace": self.namespace,
                },
                "spec": {
                    "ports": [
                        {
                            "name": "client",
                            "port": 10001,
                            "nodePort": 30001,
                        },
                    ],
                    "selector": {
                        "ray.io/identifier": f"{self.name}-head",
                    },
                    "type": "NodePort",
                },
            },
        )

    def apply(self) -> str:
        """
        Apply the Ray cluster definition to the Kubernetes cluster.

        returns the ip address of the head node
        """
        self._validate()

        api = kubernetes.client.CustomObjectsApi()

        if self._has_existing_cluster(api):
            _LOG.info(
                "Ray cluster already exists, patching instead of creating a new one."
                " This means existing nodes will not update until they are recreated."
                " Use recreate=True to force recreation."
            )
            api.patch_namespaced_custom_object(
                group="ray.io",
                version="v1",
                namespace=self.namespace,
                plural="rayclusters",
                name=self.name,
                body=self.definition,
            )
        else:
            api.create_namespaced_custom_object(
                group="ray.io",
                version="v1",
                namespace=self.namespace,
                plural="rayclusters",
                body=self.definition,
            )

        pod = self.head_node_pod

        return pod.status.pod_ip

    def delete(self) -> None:
        """
        Delete the Ray cluster from the Kubernetes cluster.
        """
        try:
            api = kubernetes.client.CustomObjectsApi()
            api.delete_namespaced_custom_object(
                group="ray.io",
                version="v1",
                namespace=self.namespace,
                plural="rayclusters",
                name=self.name,
            )
        except kubernetes.client.exceptions.ApiException as e:
            if e.status == 404:
                _LOG.info("Ray cluster does not exist, nothing to delete")
                return
            _LOG.exception("Failed to delete Ray cluster")
            raise e
        except Exception as e:
            _LOG.exception("Failed to delete Ray cluster")
            raise e

    def __enter_internal(self) -> "RayCluster":
        _LOG.info("Starting Ray cluster")
        ip = self.apply()

        if self.use_port_forward:
            warnings.warn(
                "Using port forwarding for Ray cluster"
                " is not recommended for production",
                UserWarning,
                stacklevel=3,
            )
            from geneva.runners.ray._portforward import PortForward

            self._pf = PortForward.to_head_node(self)
            self._pf.__enter__()
            ip = "localhost"
            addr = f"ray://localhost:{self._pf.local_port}"
        else:
            addr = f"ray://{ip}:10001"

        _LOG.info("Packaging local environment")
        zips = upload_local_env(zip_output_dir=self.zip_output_dir)
        geneva_zip_payload = base64.b64encode(
            json.dumps({"zips": zips}).encode("utf-8")
        ).decode("utf-8")

        _LOG.info("Connecting to Ray cluster @ %s", addr)
        try:
            ray.init(
                addr,
                runtime_env={
                    # need these two for loading the workspace deps
                    "py_modules": [geneva, pa],
                    "env_vars": {
                        "GENEVA_ZIPS": geneva_zip_payload,
                    },
                },
            )
        except Exception as e:
            _LOG.exception("Failed to connect to Ray cluster, cleaning up")
            self.delete()
            raise e

        return self

    def __enter__(self) -> "RayCluster":
        if ray.is_initialized():
            raise RuntimeError(
                "Ray is already initialized, we cannot start a new cluster"
            )

        try:
            return self.__enter_internal()
        except Exception as e:
            _LOG.exception("Failed to start Ray cluster")
            self.__exit__(None, None, None)
            raise e

    def __exit__(self, exc_type, exc_value, traceback) -> None:
        ray.shutdown()
        self.delete()

        if self.use_port_forward:
            try:
                self._pf.__exit__(exc_type, exc_value, traceback)
            except Exception:
                _LOG.exception("Failed to stop port forward")

    def _validate(self, visitor=None) -> None:
        if visitor is None:
            visitor = _ValidationVisitor()
        with (
            visitor.with_namespace(self.namespace),
            visitor.with_cluster_name(self.name),
        ):
            self.head_group._validate(visitor)
            for worker in self.worker_groups:
                worker._validate(visitor)


def _can_i(
    *,
    auth_api: kubernetes.client.AuthorizationV1Api,
    namespace: str,
    sa: str,
    verb: str,
    resource: str,
    name: str,
    group: str | None = None,
) -> bool:
    """
    Check if the service account has permission to perform the action
    """
    res = auth_api.create_namespaced_local_subject_access_review(
        namespace=namespace,
        body={
            "apiVersion": "authorization.k8s.io/v1",
            "kind": "LocalSubjectAccessReview",
            "spec": {
                "user": f"system:serviceaccount:{namespace}:{sa}",
                "resourceAttributes": {
                    "namespace": namespace,
                    "verb": verb,
                    "resource": resource,
                    "name": name,
                    **({"group": group} if group else {}),
                },
            },
        },
    )

    return res.status.allowed


@attrs.define
class _ValidationVisitor:
    cluster_name: str | None = attrs.field(init=False, default=None)
    namespace: str | None = attrs.field(init=False, default=None)

    @contextlib.contextmanager
    def with_namespace(self, namespace: str) -> Generator[None, None, None]:
        old = self.namespace
        self.namespace = namespace
        yield
        self.namespace = old

    @contextlib.contextmanager
    def with_cluster_name(self, cluster_name: str) -> Generator[None, None, None]:
        old = self.cluster_name
        self.cluster_name = cluster_name
        yield
        self.cluster_name = old

    def visit_service_account(self, sa: _ServiceAccountMixin) -> None:
        if sa.service_account is None:
            return

        # validate the service account exists
        core_api = core_v1_api.CoreV1Api()
        try:
            service_account = core_api.read_namespaced_service_account(
                name=sa.service_account, namespace=self.namespace
            )
        except kubernetes.client.exceptions.ApiException as e:
            if e.status == 404:
                raise ValueError(
                    f"Service account {sa.service_account} does not exist"
                ) from e
            raise e

        # list all the role bindings on the service account
        # and check we have the correct permissions
        auth_api = kubernetes.client.AuthorizationV1Api()
        permissions_needed = [
            {
                "verb": "get",
                "resource": "pods",
                "name": "*",
            },
            {
                "verb": "list",
                "resource": "pods",
                "name": "*",
            },
            {
                "verb": "watch",
                "resource": "pods",
                "name": "*",
            },
            {
                "verb": "get",
                "resource": "rayclusters",
                "name": self.cluster_name,
                "group": "ray.io",
            },
            {
                "verb": "patch",
                "resource": "rayclusters",
                "name": self.cluster_name,
                "group": "ray.io",
            },
        ]
        checker = functools.partial(
            _can_i,
            auth_api=auth_api,
            namespace=self.namespace,
            sa=sa.service_account,
        )
        passed = True
        error_str = ""
        for perm in permissions_needed:
            if not checker(**perm):
                error_str += (
                    f"Service account {sa.service_account} does not have the "
                    f"required permission: {perm['verb']} {perm['resource']}"
                )
                passed = False
        if not passed:
            raise ValueError(error_str)

        if service_account.metadata.annotations is None:
            raise ValueError(
                f"Service account {sa.service_account} does not have any annotations"
            )

        # TODO: need different modes of permission check here
        if "iam.gke.io/gcp-service-account" not in service_account.metadata.annotations:
            raise ValueError(
                f"Service account {sa.service_account} does not have a GCP SA"
            )

    def visit_priority_class(self, pri: _PriorityClassMixin) -> None:
        if pri.priority_class is None:
            return

        # validate the priority class exists
        scheduling_api = scheduling_v1_api.SchedulingV1Api()
        try:
            scheduling_api.read_priority_class(name=pri.priority_class)
        except kubernetes.client.exceptions.ApiException as e:
            if e.status == 404:
                raise ValueError(
                    f"Priority class {pri.priority_class} does not exist"
                ) from e
            raise e

    def visit_head_node(self, _: _HeadGroupSpec) -> None:
        pass

    def visit_worker_node(self, _: _WorkerGroupSpec) -> None:
        pass



# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: Copyright The Geneva Authors

# definition of the read task, which is portion of a fragment

import hashlib
from abc import ABC, abstractmethod
from collections.abc import Iterator
from typing import cast

import attrs
import pyarrow as pa
from typing_extensions import override

from geneva.db import connect
from geneva.query import ExtractedTransform
from geneva.table import TableReference
from geneva.transformer import UDF


class ReadTask(ABC):
    """
    A task to read data that has a defined output location and unique identifier
    """

    @abstractmethod
    def to_batches(self, *, batch_size=32) -> Iterator[pa.RecordBatch]:
        """Return the data to read"""

    @abstractmethod
    def checkpoint_key(self) -> str:
        """Return a unique key for this task"""

    @abstractmethod
    def dest_frag_id(self) -> int:
        """Return the id of the destination fragment"""

    @abstractmethod
    def dest_offset(self) -> int:
        """Return the offset into the destination fragment"""


@attrs.define(order=True)
class ScanTask(ReadTask):
    uri: str
    columns: list[str]
    frag_id: int
    offset: int
    limit: int

    version: int | None = None
    filter: str | None = None

    with_row_address: bool = False

    @override
    def to_batches(self, *, batch_size=32) -> Iterator[pa.RecordBatch]:
        uri_parts = self.uri.split("/")
        name = ".".join(uri_parts[-1].split(".")[:-1])
        db = "/".join(uri_parts[:-1])

        tbl = connect(db).open_table(name, version=self.version)
        query = tbl.search().enable_internal_api()

        if self.with_row_address:
            query = query.with_row_address()

        query = query.with_fragments(self.frag_id).offset(self.offset).limit(self.limit)
        if self.columns is not None:
            query = query.select(self.columns)
        if self.filter is not None:
            query = query.where(self.filter)

        # Currently lancedb reports the wrong type for the return value
        # of the to_batches method.  Remove pyright ignore when fixed.
        batches: pa.RecordBatchReader = query.to_batches(batch_size)  # pyright: ignore[reportAssignmentType]

        yield from batches

    @override
    def checkpoint_key(self) -> str:
        hasher = hashlib.md5()
        hasher.update(
            f"{self.uri}:{self.version}:{self.columns}:{self.frag_id}:{self.offset}:{self.limit}:{self.filter}".encode(),
        )
        return hasher.hexdigest()

    @override
    def dest_frag_id(self) -> int:
        return self.frag_id

    @override
    def dest_offset(self) -> int:
        return self.offset


@attrs.define(order=True)
class CopyTask(ReadTask):
    src: TableReference
    dst: TableReference
    columns: list[str]
    frag_id: int
    offset: int
    limit: int

    @override
    def to_batches(self, *, batch_size=32) -> Iterator[pa.RecordBatch]:
        dst_tbl = self.dst.open()
        row_ids_batch = (
            dst_tbl.search()
            .select(["__source_row_id"])
            .offset(self.offset)
            .limit(self.limit)
            .to_arrow()
        )
        row_ids = cast("list[int]", row_ids_batch["__source_row_id"].to_pylist())

        # TODO: Add streaming take to lance
        table = self.src.open().to_lance()._take_rows(row_ids, columns=self.columns)

        table = table.add_column(table.num_columns, "_rowaddr", self._get_row_addrs())

        batches = table.to_batches(max_chunksize=batch_size)

        yield from batches

    @override
    def checkpoint_key(self) -> str:
        hasher = hashlib.md5()
        hasher.update(
            f"CopyTask:{self.src.db_uri}:{self.src.table_name}:{self.src.version}:{self.columns}:{self.dst.db_uri}:{self.dst.table_name}:{self.frag_id}:{self.offset}:{self.limit}".encode(),
        )
        return hasher.hexdigest()

    @override
    def dest_frag_id(self) -> int:
        return self.frag_id

    @override
    def dest_offset(self) -> int:
        return self.offset

    def _get_row_addrs(self) -> pa.Array:
        frag_mod = self.frag_id << 32
        addrs = [frag_mod + x for x in range(self.offset, self.offset + self.limit)]
        return cast("pa.Array", pa.array(addrs, pa.uint64()))


class MapTask(ABC):
    @abstractmethod
    def checkpoint_key(self) -> str:
        """Return a unique name for the task"""

    @abstractmethod
    def name(self) -> str:
        """Return a name to use for progress strings"""

    @abstractmethod
    def apply(self, batch: pa.RecordBatch) -> pa.RecordBatch:
        """Apply the map function to the input batch, returning the output batch"""

    @abstractmethod
    def output_schema(self) -> pa.Schema:
        """Return the output schema"""

    @abstractmethod
    def is_cuda(self) -> bool:
        """Return true if the task requires CUDA"""

    @abstractmethod
    def num_cpus(self) -> float | None:
        """Return the number of CPUs the task should use (None for default)"""

    @abstractmethod
    def memory(self) -> int | None:
        """Return the amount of RAM the task should use (None for default)"""

    @abstractmethod
    def batch_size(self) -> int:
        """Return the batch size the task should use"""


@attrs.define(order=True)
class AddColumnsTask(MapTask):
    udfs: dict[str, UDF] = attrs.field()

    def __get_udf(self) -> tuple[str, UDF]:
        # TODO: Add support for multiple columns to add_columns operation
        if len(self.udfs) != 1:
            raise NotImplementedError("Add columns does not support multiple UDFs")
        return next(iter(self.udfs.items()))

    @override
    def name(self) -> str:
        name, _ = self.__get_udf()
        return name

    @override
    def checkpoint_key(self) -> str:
        _, udf = self.__get_udf()
        return udf.checkpoint_key

    @override
    def apply(self, batch: pa.RecordBatch) -> pa.RecordBatch:
        name, udf = self.__get_udf()
        new_arr = udf(batch)
        return pa.record_batch([new_arr, batch["_rowaddr"]], names=[name, "_rowaddr"])

    @override
    def output_schema(self) -> pa.Schema:
        name, udf = self.__get_udf()
        return pa.schema(
            [pa.field(name, udf.data_type), pa.field("_rowaddr", pa.uint64())]
        )

    @override
    def is_cuda(self) -> bool:
        return any(udf.cuda for udf in self.udfs.values())

    @override
    def num_cpus(self) -> float | None:
        _, udf = self.__get_udf()
        return udf.num_cpus

    @override
    def memory(self) -> int | None:
        _, udf = self.__get_udf()
        return udf.memory

    @override
    def batch_size(self) -> int:
        _, udf = self.__get_udf()
        return udf.batch_size or 32


@attrs.define(order=True)
class CopyTableTask(MapTask):
    column_udfs: list[ExtractedTransform] = attrs.field()
    view_name: str = attrs.field()
    schema: pa.Schema = attrs.field()

    @override
    def name(self) -> str:
        return self.view_name

    @override
    def checkpoint_key(self) -> str:
        return self.view_name

    @override
    def apply(self, batch: pa.RecordBatch) -> pa.RecordBatch:
        for transform in self.column_udfs:
            new_arr = transform.udf(batch)
            batch = batch.add_column(
                transform.output_index, transform.output_name, new_arr
            )
        return batch

    @override
    def output_schema(self) -> pa.Schema:
        return self.schema

    @override
    def is_cuda(self) -> bool:
        return any(column_udf.udf.cuda for column_udf in self.column_udfs)

    @override
    def num_cpus(self) -> float | None:
        return max(column_udf.udf.num_cpus for column_udf in self.column_udfs)

    @override
    def memory(self) -> int | None:
        return max(column_udf.udf.memory for column_udf in self.column_udfs)

    @override
    def batch_size(self) -> int:
        return min(column_udf.udf.batch_size or 32 for column_udf in self.column_udfs)



# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: Copyright The Geneva Authors

# Job Management

from enum import Enum, unique
from typing import TYPE_CHECKING

import attrs

if TYPE_CHECKING:
    from geneva.job.client import JobClient


@unique
class JobStatus(Enum):
    """Status of a Job"""

    PENDING = 1
    RUNNING = 2
    COMPLETED = 3
    FAILED = 4


@attrs.define
class Job:
    """A Feature Engineering Job.

    User should not directly construct this object.
    """

    db: str = attrs.field()

    table: str = attrs.field()

    job_id: str = attrs.field()

    version: int | None = attrs.field(default=None)

    status: JobStatus | None = attrs.field(init=False, default=None)

    # TODO: it is possible to generate multiple columns if the output type
    # is a struct / record batch.
    column: str | list[str] = attrs.field(init=False, default=None)
    input_columns: list[str] = attrs.field(init=False, default=None)

    # Packaging information
    docker_image: str | None = attrs.field(default=None)
    udf_uri: str | None = attrs.field(default=None)

    # Runner info
    dataflow_job_id: str | None = attrs.field(default=None, repr=False)

    # Control runtime parameters
    limit: int | None = attrs.field(default=None)
    offset: int | None = attrs.field(default=None)
    filter: str | None = attrs.field(default=None)

    _client: "JobClient" = attrs.field(init=False, default=None, repr=False)

    def __attr_post_init__(self) -> None:
        # TODO: init client, fetch job info if not provided
        pass

    def __repr__(self) -> str:
        return f"Job(id={self.job_id})"

    def start(self) -> None: ...

    def stop(self) -> None: ...



# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: Copyright The Geneva Authors

# module for packaging workspace and UDFs
# also data spec for persisting the artifacts

import abc
import base64
import hashlib
import json
import logging
from pathlib import Path

import attrs
import cloudpickle
from pyarrow.fs import FileSystem, FileType
from typing_extensions import Self

from geneva.config import ConfigBase
from geneva.packager.zip import WorkspaceZipper
from geneva.transformer import UDF

_LOG = logging.getLogger(__name__)


class UDFBackend(abc.ABC):
    @classmethod
    @abc.abstractmethod
    def packager(cls) -> "UDFPackager":
        """Return the packager for this backend."""

    def to_bytes(self) -> bytes:
        return json.dumps(attrs.asdict(self)).encode()

    @classmethod
    def from_bytes(cls, data: bytes) -> Self:
        return cls(**json.loads(data.decode()))


@attrs.define
class UDFSpec:
    """Specification for a user-defined function.

    This is an holder of an arbitrary user-defined function,
    which can use an backend for marshalling.

    The most common is likely Docker + some kind of workspace
    persistence. However, we want to support more than just
    Docker, so we create this "out most" abstraction to allow
    for more flexibility.
    """

    # the name of the udf
    name: str = attrs.field(validator=attrs.validators.min_len(1))

    # the packaging backend for the udf
    backend: str = attrs.field()

    @backend.validator
    def _validate_backend(self, attribute, value) -> None:
        backend_names = [cls.__name__ for cls in UDFBackend.__subclasses__()]
        unique_backend_names = set(backend_names)

        # ruff: noqa: ERA001
        # TODO - deduplication not working right sometimes
        # if len(backend_names) != len(unique_backend_names):
        #     duplicates = []
        #     while unique_backend_names:
        #         name = backend_names.pop()
        #         if name not in unique_backend_names:
        #             duplicates.append(name)
        #         else:
        #             unique_backend_names.discard(name)
        #     raise ValueError(
        #         "Duplicate backend names found, "
        #         "please don't define the same backend name twice."
        #         f"\nDuplicates: {duplicates}"
        #     )

        if value not in unique_backend_names:
            raise ValueError(f"Unknown backend: {value}")

    udf_payload: bytes = attrs.field()

    # the payload for the runner -- This is a HACK for allowing phalanx knowing
    # how to dispatch the UDF job. Make sure changes here are compatible to
    # parsing in phalanx.
    runner_payload: bytes | None = attrs.field(default=None)

    @classmethod
    def udf_from_spec(cls, data) -> UDF:
        # TODO: load the spec and find the backend,
        # then call the packager to do the next level unmarshalling
        pass


@attrs.define
class DockerUDFSpecV1(UDFBackend):
    """Specification for a user-defined function that runs in a Docker container.
    -- Version 1

    In this packaging spec, the python interpreter is assumed to be correctly
    setup in the container, and the user-defined function is expected to load
    using cloudpickle. With the option of downloading additional workspace
    files from a remote location (S3, GCS, etc).
    """

    # the image to run the udf in
    image: str = attrs.field()

    # the tag of the image
    tag: str | None = attrs.field()

    # spec from https://docs.docker.com/reference/cli/docker/image/tag/#description
    @tag.validator
    def _validate_tag(self, attribute, value: str | None) -> None:
        if value is None:
            return
        if len(value) > 128:
            raise ValueError("Tag must be less than 128 characters.")
        if not all(c.isalpha() or c.isnumeric or c in {"_", ".", "-"} for c in value):
            raise ValueError("Tag must be valid alphanumeric.")

    # optionally have a zip of the workspace and store it separately
    # this should be the path to the zip file on (S3, GCS, etc)
    workspace_zip: str | None = attrs.field()

    # the checksum of the workspace zip
    workspace_checksum: str | None = attrs.field()

    @workspace_checksum.validator
    def _validate_workspace_checksum(self, attribute, value) -> None:
        if self.workspace_zip and not value:
            raise ValueError(
                "Workspace checksum must not be empty when a workspace is provided."
            )

    # the udf pickle to run
    udf_pickle: bytes = attrs.field()

    @udf_pickle.validator
    def _validate_udf_pickle(self, attribute, value) -> None:
        if not value:
            raise ValueError("UDF pickle must not be empty.")
        udf = cloudpickle.loads(value)
        if not isinstance(udf, UDF):
            raise ValueError("UDF pickle must contain a UDF object.")

    # the checksum of the udf pickle
    udf_checksum: str = attrs.field(init=False)

    def __attrs_post_init__(self) -> None:
        self.udf_checksum = hashlib.sha256(self.udf_pickle).hexdigest()

    @classmethod
    def packager(cls) -> "UDFPackager":
        return DockerUDFPackager()

    def to_bytes(self) -> bytes:
        self_as_dict = attrs.asdict(self)
        self_as_dict["udf_pickle"] = base64.b64encode(
            self_as_dict["udf_pickle"]
        ).decode("utf-8")
        return json.dumps(self_as_dict).encode()

    @classmethod
    def from_bytes(cls, data: bytes) -> Self:
        self_as_dict = json.loads(data.decode())
        self_as_dict["udf_pickle"] = base64.b64decode(
            self_as_dict["udf_pickle"].encode("utf-8")
        )

        checksum = self_as_dict.pop("udf_checksum")  # not part of the init
        val = cls(**self_as_dict)
        val.udf_checksum = checksum

        return val


class UDFPackager(abc.ABC):
    """Packager for user-defined functions."""

    @abc.abstractmethod
    def marshal(self, udf: UDF) -> UDFSpec:
        """Marshal a user-defined function."""

    @abc.abstractmethod
    def unmarshal(self, spec: UDFSpec) -> UDF:
        """Unmarshal a user-defined function."""


@attrs.define
class _DockerUDFPackagerConfig(ConfigBase):
    prebuilt_docker_img: str | None = attrs.field(default=None)

    # the backend the image will eventually run on. Gets passed ot the
    # docker workspace packager it can know which base image/dockerfile
    # template to use
    runtime_backend: str | None = attrs.field(default=None)

    workspace_upload_location: str | None = attrs.field(default=None)

    @classmethod
    def name(cls) -> str:
        return "docker"


@attrs.define
class _UDFPackagerConfig(ConfigBase):
    docker: _DockerUDFPackagerConfig = attrs.field(default=_DockerUDFPackagerConfig())

    @classmethod
    def name(cls) -> str:
        return "udf"


@attrs.define
class DockerUDFPackager(UDFPackager):
    # If the user wants to use an prebuilt docker image, they can provide the
    # image name:tag here. This will be used instead of building and pushing a
    # new image.
    prebuilt_docker_img: str | None = attrs.field()

    @prebuilt_docker_img.default
    def _prebuilt_docker_img_default(self) -> str | None:
        config = _UDFPackagerConfig.get()
        if config.docker is None:
            return None
        return config.docker.prebuilt_docker_img

    # the location to upload the zipped workspace to
    # this should be the path to some directory on object storage (S3, GCS, etc)
    workspace_upload_location: str | None = attrs.field()

    @workspace_upload_location.default
    def _workspace_upload_location_default(self) -> str | None:
        config = _UDFPackagerConfig.get()
        if config.docker is None:
            return None
        return config.docker.workspace_upload_location

    # provide a zipper implementation with the correct configuration for how to
    # to zip the workspace. If workspace_zip is supplied and this is not, then
    # the packager will use the default WorkspaceZipper.
    zip_workspace_packager: WorkspaceZipper | None = attrs.field()

    @zip_workspace_packager.default
    def _zip_workspace_packager_default(self) -> WorkspaceZipper | None:
        if self.workspace_upload_location:
            return WorkspaceZipper(path=Path("."))

    def marshal(self, udf: UDF) -> UDFSpec:
        image_name, tag = "test-image", "latest"

        workspace_zip = None
        workspace_checksum = None
        if self.zip_workspace_packager:
            _LOG.info("Packaging zipped workspace")
            zip_path, checksum = self.zip_workspace_packager.zip()
            _LOG.info("Uploading zipped workspace")

            upload_location = self.workspace_upload_location
            if upload_location[-1] != "/":
                upload_location += "/"

            remote_fs, root_path = FileSystem.from_uri(upload_location)
            out_path = f"{root_path}/{checksum}.zip"
            curr_remote_file_info = remote_fs.get_file_info(out_path)
            if curr_remote_file_info.type == FileType.NotFound:
                _LOG.info(
                    f"Workspace zip does not exist, uploading {zip_path} to {out_path}"
                )
                local_fs, _local_root = FileSystem.from_uri(
                    zip_path.absolute().parent.as_uri()
                )

                with (
                    local_fs.open_input_stream(str(zip_path)) as in_file,
                    remote_fs.open_output_stream(out_path) as out_file,
                ):
                    bath_size = 1024 * 1024
                    while True:
                        buf = in_file.read(bath_size)
                        if buf:
                            out_file.write(buf)
                        else:
                            break

                _LOG.info(f"Uploaded workspace zip to {out_path}")
            else:
                _LOG.info("Workspace zip already exists, skipping upload")

            workspace_zip = f"{upload_location}{checksum}.zip"
            workspace_checksum = checksum

        udf_pickle = cloudpickle.dumps(udf)

        return UDFSpec(
            name=udf.name,
            backend=DockerUDFSpecV1.__name__,
            udf_payload=DockerUDFSpecV1(
                image=image_name,
                tag=tag,
                workspace_zip=workspace_zip,
                workspace_checksum=workspace_checksum,
                udf_pickle=udf_pickle,
            ).to_bytes(),
            runner_payload=json.dumps(
                {
                    "image": image_name + ":" + tag,
                }
            ).encode(),
        )

    def unmarshal(self, spec: UDFSpec) -> UDF:
        docker_spec = self.backend(spec)
        udf = cloudpickle.loads(docker_spec.udf_pickle)
        if not isinstance(udf, UDF):
            raise ValueError("UDF pickle must contain a UDF object.")
        return udf

    def backend(self, spec: UDFSpec) -> DockerUDFSpecV1:
        if spec.backend != DockerUDFSpecV1.__name__:
            raise ValueError("Invalid backend for UDF spec.")

        return DockerUDFSpecV1.from_bytes(spec.udf_payload)



# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: Copyright The Geneva Authors

# a debugging entrypoint for reproducing task errors locally
import argparse
import json
import logging
import pdb  # noqa: T100

import cloudpickle

from geneva import LanceCheckpointStore
from geneva.apply import ScanTask
from geneva.debug.logger import CheckpointStoreErrorLogger
from geneva.transformer import UDF

_LOG = logging.getLogger(__name__)


def run_udf_task_with_debug(
    task: ScanTask,
    udf: UDF,
) -> None:
    batches = task.to_batches()
    while True:
        try:
            batch = next(batches)
        except StopIteration:
            _LOG.info("End of task %s", task)
            break
        except Exception as e:
            _LOG.error(f"Error reading task {task}: {e}")
            pdb.post_mortem()
            break

        try:
            udf(batch)
        except Exception as e:
            _LOG.error(f"Error running UDF {udf} on batch {batch}: {e}")
            pdb.post_mortem()
            break


def register_debug_parser(subparsers: argparse._SubParsersAction) -> None:
    debug_parser = subparsers.add_parser(
        "debug", description="Run a UDF task with debugging"
    )
    debug_parser.add_argument(
        "-c",
        "--checkpoint",
        type=str,
        metavar="URI",
        help="URI to the job checkpoint store",
        required=True,
    )
    debug_parser.add_argument(
        "-j",
        "--job",
        type=str,
        metavar="ID",
        help="The id of the job to debug",
        required=True,
    )
    debug_env_group = debug_parser.add_mutually_exclusive_group(required=True)
    debug_env_group.add_argument(
        "--docker", metavar="IMAGE", help="Docker image to run the task in"
    )
    debug_env_group.add_argument(
        "--local", action="store_true", default=False, help="Run the task locally"
    )
    debug_parser.set_defaults(func=run_debug)


def run_debug(args: argparse.Namespace) -> None:
    checkpoint_store = LanceCheckpointStore(args.checkpoint_store)
    error_logger = CheckpointStoreErrorLogger(args.job_id, checkpoint_store)

    error_keys = list(error_logger.list_errors())
    for key in error_keys:
        _LOG.info(f"Found error {key}")

    while (key := input("Enter error key to debug: ")).lower() not in {
        "exit",
        "quit",
        "q",
    }:
        try:
            error_row = error_logger.get_error_row(key)
        except KeyError:
            _LOG.error(f"Error {key} not found")
            continue

        task_json = error_row["task"][0].as_py()
        udf_data = error_row["udf"][0].as_py()

        task_dict = json.loads(task_json)
        task = ScanTask(**task_dict)
        udf = cloudpickle.loads(udf_data)

        _LOG.info(f"Running task {task} with UDF {udf}")
        run_udf_task_with_debug(task, udf)



# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: Copyright The Geneva Authors

from textual.app import ComposeResult
from textual.screen import Screen
from textual.widgets import Footer, Header


class DismissOnQuitScreen(Screen):
    """
    Super simple screen that pops itself when 'q' is pressed.
    """

    BINDINGS = [
        ("q", "quit", "Quit"),
    ]

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self._result = None

    def action_quit(self) -> None:
        self.dismiss(self._result)


def nested(screen: Screen) -> ComposeResult:
    """
    A generator that yields screens from a nested screen.
    """
    for item in screen.compose():
        if isinstance(item, Footer | Header):
            continue
        yield item



# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: Copyright The Geneva Authors

import contextlib
import functools
import hashlib
import logging
import random
import uuid
from collections.abc import Iterator
from typing import cast

import attrs
import bidict
import lance
import pyarrow as pa
import ray.actor
import ray.data
import ray.exceptions
import ray.util.queue
from ray import ObjectRef

from geneva.apply import (
    CheckpointingApplier,
    plan_copy,
    plan_read,
)
from geneva.apply.applier import BatchApplier
from geneva.apply.multiprocess import MultiProcessBatchApplier
from geneva.apply.simple import SimpleApplier
from geneva.apply.task import AddColumnsTask, CopyTableTask, MapTask, ReadTask
from geneva.checkpoint import CheckpointStore
from geneva.debug.logger import CheckpointStoreErrorLogger
from geneva.job.config import JobConfig
from geneva.packager import UDFPackager
from geneva.query import (
    MATVIEW_META_BASE_DBURI,
    MATVIEW_META_BASE_TABLE,
    MATVIEW_META_BASE_VERSION,
    MATVIEW_META_QUERY,
    GenevaQuery,
    GenevaQueryBuilder,
)
from geneva.runners.ray.actor_pool import ActorPool
from geneva.runners.ray.writer import FragmentWriter
from geneva.table import TableReference
from geneva.tqdm import tqdm
from geneva.transformer import UDF

_LOG = logging.getLogger(__name__)

_SIMULATE_WRITE_FAILURE = False


@contextlib.contextmanager
def _simulate_write_failure(flag: bool) -> Iterator[None]:
    global _SIMULATE_WRITE_FAILURE
    current = _SIMULATE_WRITE_FAILURE
    try:
        _SIMULATE_WRITE_FAILURE = flag
        yield
    finally:
        _SIMULATE_WRITE_FAILURE = current


@ray.remote
@attrs.define
class ApplierActor:
    applier: CheckpointingApplier

    def __ray_ready__(self) -> None:
        pass

    def run(self, task) -> tuple[ReadTask, str]:
        return task, self.applier.run(task)


ApplierActor: ray.actor.ActorClass = cast("ray.actor.ActorClass", ApplierActor)


def _get_fragment_dedupe_key(uri: str, frag_id: int, map_task: MapTask) -> str:
    key = f"{uri}:{frag_id}:{map_task.checkpoint_key()}"
    return hashlib.sha256(key.encode()).hexdigest()


def _run_column_adding_pipeline(
    map_task: MapTask,
    checkpoint_store: CheckpointStore,
    config: JobConfig,
    dst: TableReference,
    input_plan: Iterator[ReadTask],
    job_id: str | None,
    applier_concurrency: int = 8,
    intra_applier_concurrency: int = 1,
    batch_applier: BatchApplier | None = None,
) -> None:
    plan = tqdm(input_plan)

    job_id = job_id or uuid.uuid4().hex

    if batch_applier is None:
        if intra_applier_concurrency > 1:
            batch_applier = MultiProcessBatchApplier(
                num_processes=intra_applier_concurrency
            )
        else:
            batch_applier = SimpleApplier()

    applier = CheckpointingApplier(
        map_task=map_task,
        batch_applier=batch_applier,
        checkpoint_store=checkpoint_store,
        error_logger=CheckpointStoreErrorLogger(job_id, checkpoint_store),
    )

    actor = ApplierActor

    # actor.options can only be called once, we must pass all override args
    # in one shot
    args = {
        "num_cpus": map_task.num_cpus() * intra_applier_concurrency,
    }
    if map_task.is_cuda():
        args["num_gpus"] = 1
    if map_task.memory():
        args["memory"] = map_task.memory() * intra_applier_concurrency
    actor = actor.options(**args)

    pool = ActorPool(
        functools.partial(actor.remote, applier=applier), applier_concurrency
    )

    pbar_prefix = f"[{dst.table_name} - {map_task.name()}]"

    ds = dst.open().to_lance()
    writer_pbar = tqdm(total=len(ds.get_fragments()), position=1)
    writer_pbar.set_description(f"{pbar_prefix} Writing Fragments")

    @functools.lru_cache
    def _fragment_is_done(frag_id: int) -> bool:
        dedupe_key = _get_fragment_dedupe_key(ds.uri, frag_id, map_task)
        if dedupe_key in checkpoint_store:
            writer_pbar.update(1)
            return True
        return False

    applier_iter = pool.map_unordered(
        lambda actor, value: actor.run.remote(value),
        # the API says list, but iterables are fine
        filter(lambda task: not _fragment_is_done(task.dest_frag_id()), plan),
    )

    writers = {}
    writer_queues: dict[int, ray.util.queue.Queue] = {}
    writer_futs_to_id = bidict.bidict()
    writer_task_cache = {}

    plan.set_description(f"{pbar_prefix} Applying UDFs")

    output_columns = [
        field.name for field in map_task.output_schema() if field.name != "_rowaddr"
    ]

    def _make_writer(frag_id) -> None:
        queue = ray.util.queue.Queue()
        writer = FragmentWriter.remote(
            ds.uri,
            output_columns,
            checkpoint_store,
            frag_id,
            queue,
            align_physical_rows=True,
        )
        writer_queues[frag_id] = queue
        writers[frag_id] = writer
        writer_futs_to_id[writer.write.remote()] = frag_id

    def _shutdown_writer(frag_id) -> None:
        actor_handle: ray.actor.ActorHandle = writers[frag_id]
        ray.kill(actor_handle)
        del writers[frag_id]
        writer_queue: ray.util.queue.Queue = writer_queues[frag_id]
        writer_queue.shutdown()
        del writer_queues[frag_id]
        fut = writer_futs_to_id.inverse[frag_id]
        del writer_futs_to_id[fut]

    def _restart_writer(frag_id) -> None:
        while True:
            try:
                _LOG.exception("Failed to commit fragments, restarting writer")
                _shutdown_writer(frag_id)
                _make_writer(frag_id)
                for task in writer_task_cache[frag_id]:
                    writer_queues[frag_id].put(task)
                _LOG.info("Restarted writer for fragment %d", frag_id)
                break
            # we need to keep retrying if our queue gets killed while we are
            # trying to put. There isn't a "safe_put_queue" method because we
            # could stack overflow from
            # _restart_writer -> safe_put_queue -> _restart_writer -> ...
            except (
                ray.exceptions.ActorDiedError,
                ray.exceptions.ActorUnavailableError,
            ):
                continue

    pending_fragments: list[tuple[int, lance.fragment.DataFile]] = []

    dst_read_version = ds.version

    def _do_commit(commit_granularity: int) -> None:
        nonlocal pending_fragments, dst_read_version

        if len(pending_fragments) >= commit_granularity:
            operation = lance.LanceOperation.DataReplacement(
                replacements=[
                    lance.LanceOperation.DataReplacementGroup(
                        fragment_id=frag_id,
                        new_file=new_file,
                    )
                    for frag_id, new_file in pending_fragments
                ]
            )

            lance.LanceDataset.commit(ds.uri, operation, read_version=dst_read_version)
            dst_read_version += 1
            pending_fragments = []

    def _commit(frag_id: int, fut: ObjectRef, commit_granularity: int) -> None:
        nonlocal pending_fragments
        _LOG.debug("Committing fragment id: %d", frag_id)

        try:
            fut_frag_id, new_file = ray.get(fut)
            assert fut_frag_id == frag_id
            pending_fragments.append((frag_id, new_file))
        except (ray.exceptions.ActorDiedError, ray.exceptions.ActorUnavailableError):
            _restart_writer(frag_id)
            return

        _do_commit(commit_granularity)
        _shutdown_writer(frag_id)
        del writer_task_cache[frag_id]

        writer_pbar.update(1)
        dedupe_key = _get_fragment_dedupe_key(ds.uri, frag_id, map_task)
        checkpoint_store[dedupe_key] = pa.RecordBatch.from_pydict({})

    for item in applier_iter:
        plan.set_description(
            f"{pbar_prefix} workers: {len(pool._future_to_actor)} Applying UDFs"
        )

        task: ReadTask = item[0]
        result = item[1]

        frag_id = task.dest_frag_id()
        if frag_id not in writers:
            _LOG.debug("Creating writer for fragment %d", frag_id)
            _make_writer(frag_id)
            writer_task_cache[frag_id] = []

        writer_task_cache[frag_id].append((task.dest_offset(), result))
        try:
            writer_queues[frag_id].put((task.dest_offset(), result))
        except (ray.exceptions.ActorDiedError, ray.exceptions.ActorUnavailableError):
            _restart_writer(frag_id)
        # save the task result before the fragment is committed
        # in the even the fragment writer fails we can restart it easily

        # FAULT INJECTION: simulate write failure for testing
        if _SIMULATE_WRITE_FAILURE and random.random() < 0.5:
            if random.random() < 0.5:
                ray.kill(writers[frag_id])
            else:
                ray.kill(writer_queues[frag_id].actor)

        ready, _ = ray.wait(list(writer_futs_to_id.keys()), timeout=0)
        for fut in ready:
            frag_id = writer_futs_to_id[fut]
            _commit(frag_id, fut, config.commit_granularity)

    # Force commit the remaining fragments in the buffer.
    _do_commit(1)

    # commit any remaining fragments
    # need to keep retry until all fragments are committed
    while writer_futs_to_id:
        for fut, frag_id in list(writer_futs_to_id.items()):
            # More aggressively commit restarted fragments
            _commit(frag_id, fut, 1)


def run_ray_copy_table(
    dst: TableReference,
    packager: UDFPackager,
    checkpoint_store: CheckpointStore | None = None,
    *,
    job_id: str | None = None,
    concurrency: int = 8,
    batch_size: int | None = None,
    applier_batch_size: int | None = None,
    task_shuffle_diversity: int | None = None,
    test_run: bool = True,
    commit_granularity: int | None = None,
) -> None:
    # prepare job parameters
    config = JobConfig.get().with_overrides(
        batch_size=batch_size,
        applier_batch_size=applier_batch_size,
        task_shuffle_diversity=task_shuffle_diversity,
        commit_granularity=commit_granularity,
    )

    checkpoint_store = checkpoint_store or config.make_checkpoint_store()

    dst_schema = dst.open().schema
    if dst_schema.metadata is None:
        raise Exception("Destination dataset must have view metadata.")
    src_dburi = dst_schema.metadata[MATVIEW_META_BASE_DBURI.encode("utf-8")].decode(
        "utf-8"
    )
    src_name = dst_schema.metadata[MATVIEW_META_BASE_TABLE.encode("utf-8")].decode(
        "utf-8"
    )
    src_version = int(
        dst_schema.metadata[MATVIEW_META_BASE_VERSION.encode("utf-8")].decode("utf-8")
    )
    src = TableReference(db_uri=src_dburi, table_name=src_name, version=src_version)
    query_json = dst_schema.metadata[MATVIEW_META_QUERY.encode("utf-8")]
    query = GenevaQuery.model_validate_json(query_json)

    src_table = src.open()
    schema = GenevaQueryBuilder.from_query_object(src_table, query).schema

    job_id = job_id or uuid.uuid4().hex

    column_udfs = query.extract_column_udfs(packager)

    input_cols = set()
    for column_udf in column_udfs:
        if column_udf.udf.input_columns is not None:
            for input_col in column_udf.udf.input_columns:
                input_cols.add(input_col)

    plan = plan_copy(
        src,
        dst,
        list(input_cols),
        batch_size=config.batch_size,
        task_shuffle_diversity=config.task_shuffle_diversity,
    )

    map_task = CopyTableTask(
        column_udfs=column_udfs, view_name=dst.table_name, schema=schema
    )

    _run_column_adding_pipeline(
        map_task, checkpoint_store, config, dst, plan, job_id, concurrency
    )


def run_ray_add_column(
    table: TableReference,
    columns: list[str],
    transforms: dict[str, UDF],
    checkpoint_store: CheckpointStore | None = None,
    *,
    read_version: int | None = None,
    job_id: str | None = None,
    concurrency: int = 8,
    batch_size: int | None = None,
    applier_batch_size: int | None = None,
    task_shuffle_diversity: int | None = None,
    test_run: bool = True,
    commit_granularity: int | None = None,
    **kwargs,
) -> None:
    # prepare job parameters
    config = JobConfig.get().with_overrides(
        batch_size=batch_size,
        applier_batch_size=applier_batch_size,
        task_shuffle_diversity=task_shuffle_diversity,
        commit_granularity=commit_granularity,
    )

    checkpoint_store = checkpoint_store or config.make_checkpoint_store()

    uri = table.open().to_lance().uri

    plan = plan_read(
        uri,
        columns,
        batch_size=config.batch_size,
        read_version=read_version,
        task_shuffle_diversity=config.task_shuffle_diversity,
        **kwargs,
    )

    map_task = AddColumnsTask(udfs=transforms)

    _run_column_adding_pipeline(
        map_task, checkpoint_store, config, table, plan, job_id, concurrency
    )



# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: Copyright The Geneva Authors

import pyarrow as pa
import pytest

from geneva import connect
from geneva.db import Connection
from geneva.packager import DockerUDFPackager
from geneva.query import GenevaQuery
from geneva.table import Table
from geneva.transformer import udf


@pytest.fixture
def db_and_table(tmp_path) -> (Connection, Table):
    packager = DockerUDFPackager(prebuilt_docker_img="test-image:latest")
    db = connect(tmp_path, packager=packager)
    tbl = pa.Table.from_pydict({"video_uri": ["a", "b", "c", "d", "e", "f"]})
    table = db.create_table("table", tbl)
    return db, table


def test_create_materialized_view(db_and_table) -> None:
    (db, table) = db_and_table

    @udf(data_type=pa.binary())
    def load_video(video_uri: pa.Array) -> pa.Array:
        videos = [str(i).encode("utf-8") for i in video_uri]
        return pa.array(videos)

    view_table = (
        table.search(None)
        .shuffle(seed=42)
        .select(
            {
                "video_uri": "video_uri",
                "video": load_video,
            }
        )
        .create_materialized_view(db, "table_view")
    )

    # Initially, view table is empty except for __source_row_id and __is_set
    assert view_table.to_arrow() == pa.table(
        {
            "__source_row_id": pa.array([3, 2, 5, 4, 1, 0]),
            "__is_set": pa.array([False] * 6),
            "video_uri": pa.array([None] * 6, pa.string()),
            "video": pa.array([None] * 6, pa.binary()),
        }
    )

    metadata = view_table.schema.metadata
    query = metadata[b"geneva::view::query"]

    query = GenevaQuery.model_validate_json(query)
    assert query.shuffle
    assert query.shuffle_seed == 42
    assert len(query.column_udfs) == 1
    assert query.column_udfs[0].output_name == "video"
    assert query.column_udfs[0].udf.name == "load_video"


def test_create_matview_column_not_in_select(db_and_table) -> None:
    (db, table) = db_and_table

    @udf(data_type=pa.utf8())
    def echo(video_uris: pa.Array) -> pa.Array:
        return video_uris

    view_table = (
        table.search(None)
        .select({"echo": echo})
        .create_materialized_view(db, "tbl_view")
    )

    assert view_table.to_arrow() == pa.table(
        {
            "__source_row_id": pa.array([0, 1, 2, 3, 4, 5]),
            "__is_set": pa.array([False] * 6),
            "echo": pa.array([None] * 6, pa.string()),
        }
    )



import hashlib
import random

import attrs
from kubernetes import client
from kubernetes import config as kube_config

from geneva.config import ConfigBase

KUBERAY_API_GROUP = "ray.io"
KUBERAY_API_VERSION = "v1"
KUBERAY_API_GROUP_VERSION = f"{KUBERAY_API_GROUP}/{KUBERAY_API_VERSION}"
KUBERAY_JOB_API_KIND = "RayJob"
KUBERAY_JOB_API_NAME = "rayjobs"


@attrs.define
class KuberayConfig(ConfigBase):
    checkpoint_store: str = attrs.field()
    ray_version: str = attrs.field()
    namespace: str = attrs.field(default="lancedb")
    worker_min_replicas: int = attrs.field(default=0)
    worker_max_replicas: int = attrs.field(default=10)

    @classmethod
    def name(cls) -> str:
        return "kuberay"


def launch_kuberay_job(
    db_uri: str,
    table_name: str,
    column: str,
    image: str,
    kuberay_config: KuberayConfig,
) -> None:
    # TODO api docs explaining the args supposed to be passed

    kube_config.load_kube_config()
    api = client.CustomObjectsApi()

    job_definition = {
        "apiVersion": KUBERAY_API_GROUP_VERSION,
        "kind": KUBERAY_JOB_API_KIND,
        "metadata": {
            "name": generate_job_name(db_uri, table_name, column),
            "namespace": kuberay_config.namespace,
        },
        "spec": {
            "entrypoint": f"python3 -m geneva.runners.ray --db_uri {db_uri} --table_name {table_name} --column {column} --checkpoint_store {kuberay_config.checkpoint_store}",  # noqa E501
            "rayClusterSpec": {
                "rayVersion": kuberay_config.namespace,
                "headGroupSpec": {
                    "rayStartParams": {},
                    "template": {
                        "spec": {
                            "containers": [
                                {
                                    "name": "ray-head",
                                    "image": image,
                                    "imagePullPolicy": "IfNotPresent",
                                    "ports": [
                                        {
                                            "containerPort": 10001,
                                            "name": "client",
                                            "protocol": "TCP",
                                        },
                                        {
                                            "containerPort": 8265,
                                            "name": "dashboard",
                                            "protocol": "TCP",
                                        },
                                        {
                                            "containerPort": 6379,
                                            "name": "gsc-server",
                                            "protocol": "TCP",
                                        },
                                    ],
                                }
                            ]
                        }
                    },
                },
                "workerGroupSpecs": [
                    {
                        "groupName": "worker-group-1",
                        "minReplicas": kuberay_config.worker_min_replicas,
                        "maxReplicas": kuberay_config.worker_max_replicas,
                        "rayStartParams": {},
                        "template": {
                            "spec": {
                                "containers": [
                                    {
                                        "name": "ray-worker",
                                        "image": image,
                                        "imagePullPolicy": "IfNotPresent",
                                    }
                                ]
                            }
                        },
                    }
                ],
            },
        },
    }

    api.create_namespaced_custom_object(
        group=KUBERAY_API_GROUP,
        version=KUBERAY_API_VERSION,
        namespace=kuberay_config.namespace,
        plural=KUBERAY_JOB_API_NAME,
        body=job_definition,
    )


def generate_job_name(
    db_uri: str,
    table_name: str,
    column: str,
) -> str:
    db_name = db_uri.split("/")[-1]
    seed = random.randint(0, 1000000)
    job_name = f"ray-geneva-{db_name[:6]}-{table_name[6]}-{column[:6]}-{hashlib.md5(str(seed).encode()).hexdigest()[:6]}"  # noqa E501
    return job_name



import datetime
import importlib
import random

import pytest
import ray

from geneva.runners.ray.raycluster import RayCluster, _HeadGroupSpec, _WorkerGroupSpec


def test_cluster_startup() -> None:
    cluster_name = "geneva-integ-test"
    cluster_name += f"-{datetime.datetime.now().strftime('%Y-%m-%d-%H-%M')}"
    cluster_name += f"-{random.randint(0, 10000)}"
    with RayCluster(
        name=cluster_name,
        namespace="geneva",
        use_port_forward=True,
        # allocate at least a single worker so the test runs faster
        # that we save time on waiting for the actor to start
        worker_groups=[
            _WorkerGroupSpec(
                name="worker",
                min_replicas=1,
            )
        ],
    ):
        ray.get(ray.remote(lambda: 1).remote())


def test_cluster_startup_no_gcs_permission() -> None:
    """
    Test the if we try to import geneva, which uses gcs for
    workspace packaging magic, errors when the service account
    doesn't have permission to access the gcs bucket.
    """

    cluster_name = "geneva-integ-test"
    cluster_name += f"-{datetime.datetime.now().strftime('%Y-%m-%d-%H-%M')}"
    cluster_name += f"-{random.randint(0, 10000)}"

    with (
        RayCluster(
            name=cluster_name,
            namespace="geneva",
            use_port_forward=True,
            # allocate at least a single worker so the test runs faster
            # that we save time on waiting for the actor to start
            worker_groups=[
                _WorkerGroupSpec(
                    name="worker",
                    min_replicas=1,
                )
            ],
        ),
        pytest.raises(PermissionError, match="PERMISSION_DENIED"),
    ):
        ray.get(ray.remote(lambda: importlib.import_module("geneva")).remote())


def test_cluster_startup_can_import_geneva_and_lance(
    geneva_k8s_service_account: str,
) -> None:
    """
    Test the if we try to import geneva, which uses gcs for
    workspace packaging magic, errors when the service account
    doesn't have permission to access the gcs bucket.
    """

    cluster_name = "geneva-integ-test"
    cluster_name += f"-{datetime.datetime.now().strftime('%Y-%m-%d-%H-%M')}"
    cluster_name += f"-{random.randint(0, 10000)}"

    with RayCluster(
        name=cluster_name,
        namespace="geneva",
        use_port_forward=True,
        head_group=_HeadGroupSpec(
            service_account=geneva_k8s_service_account,
        ),
        # allocate at least a single worker so the test runs faster
        # that we save time on waiting for the actor to start
        worker_groups=[
            _WorkerGroupSpec(
                name="worker",
                min_replicas=1,
                service_account=geneva_k8s_service_account,
            )
        ],
    ):
        ray.get(ray.remote(lambda: importlib.import_module("geneva")).remote())
        ray.get(ray.remote(lambda: importlib.import_module("lance")).remote())



# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: Copyright The Geneva Authors

import attrs
from typing_extensions import Self  # noqa: UP035

from geneva.checkpoint import CheckpointConfig, CheckpointStore
from geneva.config import ConfigBase


@attrs.define
class JobConfig(ConfigBase):
    """Geneva Job Configurations."""

    checkpoint: CheckpointConfig = attrs.field(default=CheckpointConfig("tempfile"))

    batch_size: int = attrs.field(default=10240, converter=int)

    task_shuffle_diversity: int = attrs.field(default=8, converter=int)

    # How many fragments to be committed in one single transaction.
    commit_granularity: int = attrs.field(default=64, converter=int)

    @classmethod
    def name(cls) -> str:
        return "job"

    def make_checkpoint_store(self) -> CheckpointStore:
        return (self.checkpoint or CheckpointConfig("tempfile")).make()

    def with_overrides(
        self,
        *,
        batch_size: int | None = None,
        applier_batch_size: int | None = None,
        task_shuffle_diversity: int | None = None,
        commit_granularity: int | None = None,
    ) -> Self:
        if batch_size is not None:
            self.batch_size = batch_size
        if applier_batch_size is not None:
            self.applier_batch_size = applier_batch_size
        if task_shuffle_diversity is not None:
            self.task_shuffle_diversity = task_shuffle_diversity
        if commit_granularity is not None:
            self.commit_granularity = commit_granularity
        return self
